"""
=======================================
Creating a CFAD diagram
=======================================
This example shows how to create a contoured frequency by altitude (CFAD) diagram
"""

print(__doc__)

# Author: Laura Tomkins (lauramtomkins@gmail.com)
# License: BSD 3 clause


import matplotlib.pyplot as plt
import numpy as np
from open_radar_data import DATASETS

import pyart

######################################
# Description of a CFAD
# ----------
# A contoured frequency by altitude diagram (CFAD) is essentially a 2D histogram used to depict the vertical
# distribution of a particular variable, such as radar reflectivity. The x-axis represents the frequency of the
# field and the y-axis represents the frequency of the altitude. A key feature that distinguishes a CFAD from a
# regular 2D histogram is that it is normalized by altitude and altitudes where there is insufficient data are
# removed. See Yuter and Houze (1995) for full details of the diagram.
#
# **Interpretation**
# In a CFAD diagram, for a given altitude, the normalized frequency values will sum to a value of 1. When
# interpreting a CFAD diagram, the normalized frequency for a given bin will be the frequency of that value for the
# associated altitude (i.e. each bin can be interpreted as the fraction of data points at each altitude).
#
# **Minimum fraction threshold**
# In a CFAD diagram, altitudes where there is insufficient data are removed. This feature is controlled with the
# `min_frac_thres` value. The default value is 0.1, meaning that altitudes where the number of observations is less
# than one tenth of the maximum number of observations across all altitudes are removed. Increasing the value will
# act to be more aggressive in removing altitudes and decreasing the value will act to include more data. Setting
# this value to zero will include all altitudes with data, but use caution when interpreting CFADs with all data
# available as those altitudes with less data will not be as representative as other altitudes with sufficient data.
######################################
# Basic example with reflectivity from RHI
# ----------
# **Example with RHI**
# First, we will show an example from an RHI scan. We will get the reflectivity data and mask outside -15 and 35 dBZ
# to remove any noisy data. For the best results, we recommend cleaning up the field as much as possible (i.e.
# despeckling, filtering, etc.)

# get test data
filename = pyart.testing.get_test_data("sgpxsaprrhicmacI5.c0.20110524.015604_NC4.nc")
radar = pyart.io.read_cfradial(filename)

# compute CFAD
# get reflectivity data and mask extremes
# extract first sweep
radar = radar.extract_sweeps([0])
# get mask array
ref_data = radar.fields["reflectivity_horizontal"]["data"][:]
ref_data_masked = np.ma.masked_outside(ref_data, -15, 35)
field_mask = ref_data_masked.mask

# get CFAD
freq_norm, height_edges, field_edges = pyart.retrieve.create_cfad(
    radar,
    field_bins=np.linspace(-15, 35, 100),
    altitude_bins=np.linspace(0, 15000, 50),
    field="reflectivity_horizontal",
    field_mask=field_mask,
    min_frac_thres=0.1,
)

# plot CFAD
# mask frequency values less than zero
freq_norm_masked = np.ma.masked_less_equal(freq_norm, 0)
# plot CFAD
plt.figure()
ax = plt.axes()
cfad_pm = ax.pcolormesh(
    field_edges, height_edges / 1000, freq_norm_masked, cmap="plasma", vmin=0, vmax=0.10
)
plt.colorbar(cfad_pm)
ax.set_xlabel("Reflectivity [dBZ]")
ax.set_ylabel("Height [km]")
ax.grid(ls="--", color="gray", lw=0.5, alpha=0.7)
plt.show()

# plot RHI data
display = pyart.graph.RadarDisplay(radar)

plt.figure(figsize=[12, 3])
ax = plt.axes()
plt.tight_layout()
display.plot(
    "reflectivity_horizontal",
    0,
    vmin=-15,
    vmax=35,
    mask_outside=True,
    cmap="magma_r",
    ax=ax,
)
display.set_limits(ylim=[0, 15], ax=ax)
ax.set_aspect("equal")
plt.show()

######################################
# We can see a general increase in reflectivity values from the echo top to around 8km. The maximum frequency value in
# each altitude row represents the mode of the reflectivity distribution which we can see also increases from echo
# top to 8km. Below 8km the distribution of reflectivity values widens, likely associated with some of the noise in
# the RHI.

# Minimum fraction threshold example
# ----------
# Previously, we used the default `min_frac_thres` of 0.1. Next, we will increase the threshold and set the threshold
# to 0 (not recommended) so show how the CFAD changes.

# Let's see the effect of changing the minimum fraction threshold:
freq_norm2, height_edges, field_edges = pyart.retrieve.create_cfad(
    radar,
    field_bins=np.linspace(-15, 35, 100),
    altitude_bins=np.linspace(0, 15000, 50),
    field="reflectivity_horizontal",
    field_mask=field_mask,
    min_frac_thres=0.2,
)
freq_norm0, height_edges, field_edges = pyart.retrieve.create_cfad(
    radar,
    field_bins=np.linspace(-15, 35, 100),
    altitude_bins=np.linspace(0, 15000, 50),
    field="reflectivity_horizontal",
    field_mask=field_mask,
    min_frac_thres=0,
)

# plot CFAD
# mask zero values
freq_norm2_masked = np.ma.masked_less_equal(freq_norm2, 0)
freq_norm0_masked = np.ma.masked_less_equal(freq_norm0, 0)
# plot
plt.figure(figsize=(12, 3))
ax1 = plt.subplot(1, 3, 3)
cfad_pm = ax1.pcolormesh(
    field_edges,
    height_edges / 1000,
    freq_norm2_masked,
    cmap="plasma",
    vmin=0,
    vmax=0.10,
)
plt.colorbar(cfad_pm, ax=ax1)
ax1.set_xlabel("Reflectivity [dBZ]")
ax1.grid(ls="--", color="gray", lw=0.5, alpha=0.7)
ax1.set_title("min_frac_thres = 0.2")

ax2 = plt.subplot(1, 3, 2)
cfad_pm = ax2.pcolormesh(
    field_edges,
    height_edges / 1000,
    freq_norm_masked,
    cmap="plasma",
    vmin=0,
    vmax=0.10,
)
plt.colorbar(cfad_pm, ax=ax2)
ax2.set_xlabel("Reflectivity [dBZ]")
ax2.grid(ls="--", color="gray", lw=0.5, alpha=0.7)
ax2.set_title("min_frac_thres = 0.1")

ax3 = plt.subplot(1, 3, 1)
cfad_pm = ax3.pcolormesh(
    field_edges,
    height_edges / 1000,
    freq_norm0_masked,
    cmap="plasma",
    vmin=0,
    vmax=0.10,
)
plt.colorbar(cfad_pm, ax=ax3)
ax3.set_xlabel("Reflectivity [dBZ]")
ax3.set_ylabel("Height [km]")
ax3.grid(ls="--", color="gray", lw=0.5, alpha=0.7)
ax3.set_title("min_frac_thres = 0")
plt.show()

######################################
# Setting the `min_frac_thres` to 0 (left panel) shows all data, even near the top of the chart (14km) where there is
# limited echo. Setting the `min_frac_thres` higher to 0.2 (right panel) removed altitudes between 1 and 5 km where
# there is less echo than between 6 and 12km where there is a consistent swath of reflectivity throughout the entire
# cross section.
#
#
# Velocity example
# ----------
# Next, we will show a CFAD for the doppler velocity from the above example. First, we have to dealias the velocity.

# create a gatefilter
gatefilter = pyart.filters.GateFilter(radar)
gatefilter.exclude_invalid("reflectivity_horizontal")
gatefilter.exclude_outside("reflectivity_horizontal", -15, 30)
gatefilter.exclude_invalid("mean_doppler_velocity")

# dealias velocity
velocity_dealias = pyart.correct.dealias_region_based(
    radar,
    gatefilter=gatefilter,
    vel_field="mean_doppler_velocity",
    nyquist_vel=13,
)
radar.add_field("corrected_velocity", velocity_dealias)

# plot RHI data
display = pyart.graph.RadarDisplay(radar)

plt.figure(figsize=[12, 3])
ax = plt.axes()
plt.tight_layout()
display.plot(
    "corrected_velocity",
    0,
    vmin=-20,
    vmax=20,
    mask_outside=True,
    cmap="RdBu_r",
    ax=ax,
)
display.set_limits(ylim=[0, 15], ax=ax)
ax.set_aspect("equal")
plt.show()

freq_norm, height_edges, field_edges = pyart.retrieve.create_cfad(
    radar,
    field_bins=np.linspace(-30, 30, 50),
    altitude_bins=np.linspace(0, 15000, 50),
    field="corrected_velocity",
    field_mask=field_mask,
    min_frac_thres=0.2,
)

# plot CFAD
# mask zero values
freq_norm_masked = np.ma.masked_less_equal(freq_norm, 0)

# plot
plt.figure()
ax = plt.axes()
cfad_pm = ax.pcolormesh(
    field_edges,
    height_edges / 1000,
    freq_norm_masked,
    cmap="plasma",
    vmin=0,
    vmax=0.20,
)
plt.colorbar(cfad_pm)
ax.set_xlabel("Velocity [m s$^{-1}$]")
ax.set_ylabel("Height [km]")
ax.grid(ls="--", color="gray", lw=0.5, alpha=0.7)
plt.show()

######################################
# The velocity CFAD is very different from the reflectivity CFAD. In most altitudes, there is more of a bimodal
# pattern associated with the changing sign of the velocity values on either side of the radar. In general,
# the distribution of velocity values is consistently wide throughout the profile compared to the reflectivity CFAD.
#
# Validation
# ----------
# Finally, we wanted to compare this function with the original method, so here we reproduce Fig. 2c from Yuter and
# Houze (1995) to demonstrate that it works the same. Instead of using the `pcolormesh` function, we are using
# contour lines.


# get test data
filename = DATASETS.fetch("ddop.910815.213931.cdf")
grid = pyart.io.read_grid(filename)

# make corrections to altitude field
altitude_data = grid.point_z["data"]
grid.point_z["data"] = (altitude_data - 800) / 1000

# get fields to create a mask
field_data = grid.fields["maxdz"]["data"][:]
vvel_data = grid.fields["w_wind"]["data"][:]
vvel_masked = np.ma.masked_invalid(vvel_data)
field_data_masked = np.ma.masked_less_equal(field_data, -15)
field_data_masked = np.ma.masked_where(vvel_masked.mask, field_data_masked)

# define histogram bins
field_bins = np.arange(-20, 65, 5)
altitude_bins = np.arange(-0.2, 18.5, 0.4)

freq_norm, height_edges, field_edges = pyart.retrieve.create_cfad(
    grid,
    field_bins=field_bins,
    altitude_bins=altitude_bins,
    field="maxdz",
    field_mask=field_data_masked.mask,
    min_frac_thres=0.1,
)

# plot CFAD with contour plot
freq_norm_masked = np.ma.masked_less_equal(freq_norm, 0)
h, f = np.meshgrid(height_edges, field_edges)

plt.figure(figsize=(6, 3))
ax = plt.axes()
cont = ax.contour(
    f[:-1, :-1] + 2.5,
    h[:-1, :-1] + 0.2,
    freq_norm.T,
    levels=np.arange(0.05, 0.3, 0.05),
    colors="black",
)
ax.set_yticks([0, 5, 10, 15])
ax.set_xticks([-10, 0, 10, 20, 30, 40, 50])
ax.set_xlim([-12, 58])
ax.set_ylim([-0.5, 16])
ax.set_ylabel("Height [km]")
ax.set_xlabel("Reflectivity [dBZ]")
ax.axhline(8, ls="--", lw=0.75, color="black")
ax.set_title("Yuter and Houze (1995) Fig. 2c")
plt.show()

# References
# ----------
# Yuter, S. E., and R. A. Houze, 1995: Three-Dimensional Kinematic and Microphysical Evolution of Florida
# Cumulonimbus. Part II: Frequency Distributions of Vertical Velocity, Reflectivity, and Differential Reflectivity.
# Mon. Wea. Rev. 123, 1941-1963. https://doi.org/10.1175/1520-0493(1995)123%3C1941:TDKAME%3E2.0.CO;2



""" Unit Tests for Py-ART's retrieve/echo_class.py module. """

import numpy as np
import pytest
from numpy.testing import assert_allclose

import pyart


def test_steiner_conv_strat_default():
    grid = pyart.testing.make_storm_grid()
    eclass_default = pyart.retrieve.steiner_conv_strat(grid)
    assert np.all(
        eclass_default["data"][25]
        == np.array(
            [
                0,
                0,
                0,
                0,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                0,
                0,
                0,
                0,
            ]
        )
    )


@pytest.mark.parametrize("area_relation", ["small", "medium", "large", "sgp"])
def test_steiner_conv_strat_modify_area(area_relation):
    grid = pyart.testing.make_storm_grid()
    eclass = pyart.retrieve.steiner_conv_strat(grid, area_relation=area_relation)
    assert eclass["data"].min() == 0
    assert eclass["data"].max() == 2


def test_feature_detection_default():
    grid = pyart.testing.make_storm_grid()
    dict = pyart.retrieve.feature_detection(grid, bkg_rad_km=50)

    assert "feature_detection" in dict.keys()
    assert "feature_under" in dict.keys()
    assert "feature_over" in dict.keys()
    assert np.all(
        dict["feature_detection"]["data"][0, 25, :]
        == np.array(
            [
                0,
                0,
                0,
                0,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                2,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                0,
                0,
                0,
                0,
            ]
        )
    )


def test_feature_detection_noest():
    grid = pyart.testing.make_storm_grid()
    dict = pyart.retrieve.feature_detection(grid, bkg_rad_km=50, estimate_flag=False)

    assert "feature_detection" in dict.keys()
    assert "feature_under" not in dict.keys()
    assert "feature_over" not in dict.keys()


def test_hydroclass_semisupervised():
    radar = pyart.io.read(pyart.testing.NEXRAD_ARCHIVE_MSG31_FILE)
    temp_dict = pyart.config.get_metadata("temperature")
    temp = np.empty(radar.fields["reflectivity"]["data"].shape)
    temp[:] = -10.0
    temp_dict["data"] = temp
    radar.add_field("temperature", temp_dict, replace_existing=True)
    radar.fields["specific_differential_phase"] = radar.fields.pop("differential_phase")
    radar_small = radar.extract_sweeps([0])

    mass_centers = pyart.retrieve.echo_class._get_mass_centers(10e9)

    hydro_nofreq = pyart.retrieve.hydroclass_semisupervised(radar_small)["hydro"]
    assert "units" in hydro_nofreq.keys()
    assert "standard_name" in hydro_nofreq.keys()
    assert "long_name" in hydro_nofreq.keys()
    assert "coordinates" in hydro_nofreq.keys()

    assert hydro_nofreq["data"].dtype == "uint8"
    assert_allclose(hydro_nofreq["data"][0][0:5], [6, 6, 6, 6, 6])
    assert_allclose(hydro_nofreq["data"][0][-5:], [2, 2, 2, 2, 2])

    radar_small.instrument_parameters["frequency"] = {"data": np.array([5e9])}
    hydro_freq = pyart.retrieve.hydroclass_semisupervised(radar_small)["hydro"]
    assert_allclose(hydro_freq["data"][0][0:5], [6, 6, 6, 6, 6])
    assert_allclose(hydro_freq["data"][0][-5:], [2, 2, 2, 2, 2])

    hydro = pyart.retrieve.hydroclass_semisupervised(
        radar_small,
        mass_centers=mass_centers,
        compute_entropy=True,
        output_distances=True,
    )
    assert_allclose(hydro["hydro"]["data"][0][0:5], [6, 6, 6, 6, 6])
    assert_allclose(hydro["hydro"]["data"][0][-5:], [2, 2, 2, 2, 2])
    assert_allclose(
        hydro["entropy"]["data"][0][0:5].data,
        [0.35945634, 0.35945634, 0.35945634, 0.35945634, 0.35945634],
        rtol=1e-5,
    )
    assert_allclose(
        hydro["entropy"]["data"][0][-5:].data,
        [0.232788, 0.232788, 0.232788, 0.232788, 0.232788],
        rtol=1e-5,
    )
    assert_allclose(
        hydro["proportion_CR"]["data"][0][0:5].data,
        [0.03524214, 0.03524214, 0.03524214, 0.03524214, 0.03524214],
        rtol=1e-5,
    )
    assert_allclose(
        hydro["proportion_CR"]["data"][0][-5:].data,
        [
            8.336829723993246e-05,
            8.336829723993246e-05,
            8.336829723993246e-05,
            8.336829723993246e-05,
            8.336829723993246e-05,
        ],
        rtol=1e-5,
    )


def test_data_limits_table():
    dlimits_dict = pyart.retrieve.echo_class._data_limits_table()
    test_dict = {
        "Zh": (60.0, -10.0),
        "ZDR": (5, -1.5),
        "KDP": (7.0, -10.0),
        "RhoHV": (-5.23, -50.0),
        "RelH": (5000, -5000),
    }
    assert isinstance(dlimits_dict, dict)
    assert dlimits_dict == test_dict


def test_get_freq():
    freq_band_s = pyart.retrieve.get_freq_band(3e9)
    assert freq_band_s == "S"

    freq_band_c = pyart.retrieve.get_freq_band(6e9)
    assert freq_band_c == "C"

    freq_band_x = pyart.retrieve.get_freq_band(10e9)
    assert freq_band_x == "X"

    freq_band_bad = pyart.retrieve.get_freq_band(10)
    assert freq_band_bad is None


def test_mass_centers_dict():
    mass_centers_dict = pyart.retrieve.echo_class._mass_centers_table()

    assert type(mass_centers_dict)
    assert "C" in mass_centers_dict.keys()
    assert "X" in mass_centers_dict.keys()

    assert_allclose(
        mass_centers_dict["C"][0],
        [1.35829e01, 4.06300e-01, 4.97000e-02, 9.86800e-01, 1.33030e03],
        atol=1e-7,
    )
    assert_allclose(
        mass_centers_dict["C"][8],
        [5.06186e01, -6.49000e-02, 9.46000e-02, 9.90400e-01, 1.17990e03],
        atol=1e-7,
    )

    assert_allclose(
        mass_centers_dict["X"][0],
        [1.9077e01, 4.1390e-01, 9.9000e-03, 9.8410e-01, 1.0617e03],
        atol=1e-7,
    )

    assert_allclose(
        mass_centers_dict["X"][8],
        [4.42216e01, -3.41900e-01, 6.87000e-02, 9.68300e-01, 1.27270e03],
        atol=1e-7,
    )


def test_mass_centers():
    mass_centers_good_freq = pyart.retrieve.echo_class._get_mass_centers(10e9)
    assert_allclose(
        mass_centers_good_freq[0],
        [1.90770e01, 4.13900e-01, 9.90000e-03, 9.84100e-01, 1.06170e03],
        atol=1e-7,
    )

    low_bad_freq = pyart.retrieve.echo_class._get_mass_centers(3e9)
    assert_allclose(
        low_bad_freq[0],
        [1.35829e01, 4.06300e-01, 4.97000e-02, 9.86800e-01, 1.33030e03],
        atol=1e-7,
    )

    high_bad_freq = pyart.retrieve.echo_class._get_mass_centers(13e9)
    assert_allclose(
        high_bad_freq[0],
        [1.90770e01, 4.13900e-01, 9.90000e-03, 9.84100e-01, 1.06170e03],
        atol=1e-7,
    )


def test_assign_to_class():
    rhohv = np.array(([0.2, 0.2, 0.2, 0.2, 0.2], [0.2, 0.2, 0.2, 0.2, 0.2]))
    zdr = np.array(([-7.8, -7.8, -7.8, -7.8, -7.8], [-7.8, -7.8, -7.8, -7.8, -7.8]))
    kdp = np.array(
        ([180.5, 180.5, 180.5, 180.5, 180.5], [180.5, 180.5, 180.5, 180.5, 180.5])
    )
    relh = np.array(([10.0, 10.0, 10.0, 10.0, 10.0], [10.0, 10.0, 10.0, 10.0, 10.0]))
    zh = np.array(
        ([-32.0, -32.0, -32.0, -32.0, -32.0], [-32.0, -32.0, -32.0, -32.0, -32.0])
    )

    mass_centers = pyart.retrieve.echo_class._get_mass_centers(10e9)
    field_dict = {"Zh": zh, "ZDR": zdr, "KDP": kdp, "RhoHV": rhohv, "relH": relh}

    hydroclass, _, _ = pyart.retrieve.echo_class._assign_to_class(
        field_dict, mass_centers
    )
    assert_allclose(hydroclass[0], [7, 7, 7, 7, 7], atol=1e-7)


def test_standardize():
    kdp = np.array(
        ([180.5, 180.5, 180.5, 180.5, 180.5], [180.5, 180.5, 180.5, 180.5, 180.5])
    )
    field_std_kdp = pyart.retrieve.echo_class._standardize(
        kdp, "KDP", mx=180.0, mn=181.0
    )
    assert_allclose(field_std_kdp[0], [-1.0, -1.0, -1.0, -1.0, -1.0], atol=1e-6)

    relh = np.array(([10.0, 10.0, 10.0, 10.0, 10.0], [10.0, 10.0, 10.0, 10.0, 10.0]))
    field_std_relh = pyart.retrieve.echo_class._standardize(
        relh, "relH", mx=9.0, mn=5.0
    )
    assert_allclose(
        field_std_relh[0], [0.024994, 0.024994, 0.024994, 0.024994, 0.024994], atol=1e-6
    )

    rhohv = np.array(([0.2, 0.2, 0.2, 0.2, 0.2], [0.2, 0.2, 0.2, 0.2, 0.2]))
    field_std_rhohv = pyart.retrieve.echo_class._standardize(
        rhohv, "RhoHV", mx=0.1, mn=1.0
    )
    assert_allclose(field_std_rhohv[0], [-1.0, -1.0, -1.0, -1.0, -1.0], atol=1e-6)

    field_std_no_limits = pyart.retrieve.echo_class._standardize(
        rhohv, "RhoHV", mx=None, mn=None
    )
    assert_allclose(field_std_no_limits[0], [1.0, 1.0, 1.0, 1.0, 1.0], atol=1e-6)

    pytest.raises(ValueError, pyart.retrieve.echo_class._standardize, rhohv, "foo")


def test_conv_strat_raut_outDict_valid():
    """
    Test that function returns a valid dictionary with all expected keys'.
    """

    # Test that function raises `TypeError` with invalid grid object as input.
    pytest.raises(TypeError, pyart.retrieve.conv_strat_raut, None, "foo")

    # Create a Gaussian storm grid
    grid_len = 32
    gaussian_storm_2d = pyart.testing.make_gaussian_storm_grid(
        min_value=5, max_value=45, grid_len=grid_len, sigma=0.2, mu=0, masked_boundary=3
    )
    wtclass = pyart.retrieve.conv_strat_raut(
        gaussian_storm_2d, "reflectivity", cappi_level=0
    )

    # First check that it's a Python dictionary
    assert isinstance(wtclass, dict), "Output is not a dictionary"
    # then test 'wt_reclass' key exists in the dict
    assert "wt_reclass" in wtclass.keys()

    # Now test other expected keys
    expected_keys = [
        "data",
        "standard_name",
        "long_name",
        "valid_min",
        "valid_max",
        "classification_description",
        "parameters",
    ]

    # Get the keys of the 'wt_reclass' field
    reclass_keys = wtclass["wt_reclass"].keys()

    # Check each expected key
    for key in expected_keys:
        assert key in reclass_keys

    # check grid shape
    assert wtclass["wt_reclass"]["data"].shape == (1, grid_len, grid_len)


def test_conv_strat_raut_results_correct():
    """
    Checks the correctness of the results from the function.

    I created a fixed Gaussian storm with masked boundaries as pyart grid and classified it.
    Then constructed manually the expected classification results and compared it to the actual output from the function.
    """

    # Create a Gaussian storm grid
    grid_len = 32
    mask_margin = 3

    gaussian_storm_2d = pyart.testing.make_gaussian_storm_grid(
        min_value=5,
        max_value=45,
        grid_len=grid_len,
        sigma=0.2,
        mu=0,
        masked_boundary=mask_margin,
    )

    wtclass = pyart.retrieve.conv_strat_raut(gaussian_storm_2d, "reflectivity")

    # Create a 32x32 array of ones
    test_reclass = np.ones((grid_len, grid_len))

    # Mask the edges
    test_reclass[:mask_margin, :] = np.nan
    test_reclass[-mask_margin:, :] = np.nan
    test_reclass[:, :mask_margin] = np.nan
    test_reclass[:, -mask_margin:] = np.nan

    # Define the center and create the 4x4 area
    center = grid_len // 2
    # these are actual results from successful run
    test_reclass[center - 3 : center + 3, center - 3 : center + 3] = 2
    test_reclass[center - 2 : center + 2, center - 2 : center + 2] = 3

    test_reclass[13, 13] = 1
    test_reclass[13, 18] = 1
    test_reclass[18, 13] = 1
    test_reclass[18, 18] = 1

    # Creating a mask for NaN values
    mask = np.isnan(test_reclass)
    masked_reclass = np.ma.array(test_reclass, mask=mask).astype(np.int32)
    masked_reclass = np.expand_dims(masked_reclass, axis=0)

    assert_allclose(masked_reclass, wtclass["wt_reclass"]["data"], atol=0.1)



"""
=================================================
Dealias Radial Velocities Using Xradar and Py-ART
=================================================

An example which uses xradar and Py-ART to dealias radial velocities.

"""

# Author: Max Grover (mgrover@anl.gov)
# License: BSD 3 clause


import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import xradar as xd

import pyart
from pyart.testing import get_test_data

# Locate the test data and read in using xradar
filename = get_test_data("swx_20120520_0641.nc")
tree = xd.io.open_cfradial1_datatree(filename)

# Give the tree Py-ART radar methods
radar = tree.pyart.to_radar()

# Determine the nyquist velocity using the maximum radial velocity from the first sweep
nyq = radar["sweep_0"]["mean_doppler_velocity"].max().values

# Set the nyquist to what we captured above
# Calculate the velocity texture
vel_texture = pyart.retrieve.calculate_velocity_texture(
    radar, vel_field="mean_doppler_velocity", nyq=nyq
)
radar.add_field("velocity_texture", vel_texture, replace_existing=True)

##########################################
# **Visualize our velocity texture field**
# Let's use the RadarMapDisplay to visualize the texture field

fig = plt.figure(figsize=[8, 8])
display = pyart.graph.RadarMapDisplay(radar)
display.plot_ppi_map(
    "velocity_texture",
    sweep=2,
    resolution="50m",
    vmin=0,
    vmax=10,
    projection=ccrs.PlateCarree(),
    cmap="balance",
)
plt.show()

##################################################
# **Determine a Velocity Texture Threshold Value**
#
# We can use the xradar/xarray plotting functionality here
radar["sweep_0"]["velocity_texture"].plot.hist()
plt.show()

#############################
# **Apply a Gatefilter Mask**
#
# We now apply this threshold, along with a reflectivity threshold,
# and make use of the region-based dealiasing algorithm

# Configure the gatefilter
gatefilter = pyart.filters.GateFilter(radar)
gatefilter.exclude_above("velocity_texture", 4)
gatefilter.exclude_below("corrected_reflectivity_horizontal", 0)

# At this point, we can simply used dealias_region_based to dealias the velocities
# and then add the new field to the radar.
velocity_dealiased = pyart.correct.dealias_region_based(
    radar,
    vel_field="mean_doppler_velocity",
    nyquist_vel=nyq,
    centered=True,
    gatefilter=gatefilter,
)
radar.add_field("corrected_velocity", velocity_dealiased, replace_existing=True)

#############################################
# **Visualize the Cleaned Radial Velocities**
#
# We can visualize the uncorrected and corrected radial velocity fields
fig = plt.figure(figsize=(14, 5))
display = pyart.graph.RadarMapDisplay(radar)
ax1 = plt.subplot(121)
display.plot_ppi(
    "mean_doppler_velocity",
    cmap="twilight_shifted",
    vmin=-40,
    vmax=40,
    colorbar_label="Uncorrected Radial Velocity (m/s)",
    ax=ax1,
)
ax2 = plt.subplot(122)
display.plot_ppi(
    "corrected_velocity",
    cmap="twilight_shifted",
    vmin=-40,
    vmax=40,
    colorbar_label="Corrected Radial Velocity (m/s)",
    ax=ax2,
)
plt.show()



"""
==================================
Plot a PPI Using Xradar and Py-ART
==================================

An example which uses xradar and Py-ART to create a PPI plot of a Cfradial file.

"""

# Author: Max Grover (mgrover@anl.gov)
# License: BSD 3 clause


import xradar as xd

import pyart
from pyart.testing import get_test_data

# Locate the test data and read in using xradar
filename = get_test_data("swx_20120520_0641.nc")
tree = xd.io.open_cfradial1_datatree(filename)

# Give the tree Py-ART radar methods
radar = tree.pyart.to_radar()

# Plot the Reflectivity Field (corrected_reflectivity_horizontal)
display = pyart.graph.RadarMapDisplay(radar)
display.plot_ppi(
    "corrected_reflectivity_horizontal", cmap="ChaseSpectral", vmin=-20, vmax=70
)



""" Unit Tests for Py-ART's correct/proc_phase.py module. """

# Can also be run as a script to create a ray_plot.png file
# python test_phase_proc.py
# This file should look like the reference_ray_plot.png file

# Use:
# python test_phase_proc.py -r
# to recreate the reference_rays.npz and reference_ray_plot.png files

import os
import warnings

import numpy as np
import pytest

try:
    import cvxopt  # noqa

    cvxopt_available = True
except ImportError:
    cvxopt_available = False

try:
    import glpk  # noqa

    glpk_available = True
except ImportError:
    glpk_available = False


try:
    import cylp.cy  # noqa

    cylp_available = True
except ImportError:
    cylp_available = False

import pyart

PATH = os.path.dirname(__file__)
REFERENCE_RAYS_FILE = os.path.join(PATH, "reference_rays.npz")


@pytest.mark.skipif(not glpk_available, reason="GLPK is not installed.")
def test_phase_proc_lp_glpk():
    radar, phidp, kdp = perform_phase_processing()
    ref = np.load(REFERENCE_RAYS_FILE)
    assert _ratio(ref["reference_phidp"], phidp["data"]) <= 0.01
    assert _ratio(ref["reference_kdp"], kdp["data"]) <= 0.01
    assert (
        _ratio(
            ref["reference_unfolded_phidp"],
            radar.fields["unfolded_differential_phase"]["data"],
        )
        <= 0.01
    )


@pytest.mark.skipif(not cvxopt_available, reason="CVXOPT is not installed.")
def test_phase_proc_lp_cvxopt():
    from cvxopt import solvers

    solvers.options["LPX_K_MSGLEV"] = 0  # supress screen output
    radar, phidp, kdp = perform_phase_processing("cvxopt")
    ref = np.load(REFERENCE_RAYS_FILE)
    assert _ratio(ref["reference_phidp"], phidp["data"]) <= 0.01
    assert _ratio(ref["reference_kdp"], kdp["data"]) <= 0.01
    assert (
        _ratio(
            ref["reference_unfolded_phidp"],
            radar.fields["unfolded_differential_phase"]["data"],
        )
        <= 0.01
    )


@pytest.mark.skipif(not cylp_available, reason="CyLP is not installed.")
def test_phase_proc_lp_cylp():
    with warnings.catch_warnings():
        # ignore FutureWarnings as CyLP emits a number of these
        warnings.simplefilter("ignore", category=FutureWarning)
        radar, phidp, kdp = perform_phase_processing("cylp")
    ref = np.load(REFERENCE_RAYS_FILE)
    assert _ratio(ref["reference_phidp"], phidp["data"]) <= 0.01
    assert _ratio(ref["reference_kdp"], kdp["data"]) <= 0.01
    assert (
        _ratio(
            ref["reference_unfolded_phidp"],
            radar.fields["unfolded_differential_phase"]["data"],
        )
        <= 0.01
    )


@pytest.mark.skipif(not cylp_available, reason="CyLP is not installed.")
def test_phase_proc_lp_gf_cylp():
    with warnings.catch_warnings():
        # ignore FutureWarnings as CyLP emits a number of these
        warnings.simplefilter("ignore", category=FutureWarning)
        radar, phidp, kdp = test_perform_phase_processing_gf("cylp")
    ref = np.load(REFERENCE_RAYS_FILE)
    assert _ratio(ref["reference_phidp"], phidp["data"]) <= 0.05
    assert _ratio(ref["reference_kdp"], kdp["data"]) <= 0.05
    assert (
        _ratio(
            ref["reference_unfolded_phidp"],
            radar.fields["unfolded_differential_phase"]["data"],
        )
        <= 0.05
    )


@pytest.mark.skipif(not cylp_available, reason="CyLP is not installed.")
def test_phase_proc_lp_cylp_mp():
    with warnings.catch_warnings():
        # ignore FutureWarnings as CyLP emits a number of these
        warnings.simplefilter("ignore", category=FutureWarning)
        radar, phidp, kdp = perform_phase_processing("cylp_mp")
    ref = np.load(REFERENCE_RAYS_FILE)
    assert _ratio(ref["reference_phidp"], phidp["data"]) <= 0.01
    assert _ratio(ref["reference_kdp"], kdp["data"]) <= 0.01
    assert (
        _ratio(
            ref["reference_unfolded_phidp"],
            radar.fields["unfolded_differential_phase"]["data"],
        )
        <= 0.01
    )


def _ratio(a1, a2):
    """Ratio the sum of the abs difference vs sum abs of two vectors."""
    abs_residues = np.abs(a1 - a2).sum()
    avg_abs_sum = 0.5 * np.abs(a1).sum() + 0.5 * np.abs(a2).sum()
    return abs_residues / avg_abs_sum


def perform_phase_processing(LP_solver="cvxopt"):
    """Perform LP phase processing on a single ray radar."""
    radar = pyart.testing.make_single_ray_radar()
    phidp, kdp = pyart.correct.phase_proc_lp(radar, 0.0, LP_solver=LP_solver)
    return radar, phidp, kdp


@pytest.mark.skipif(not cvxopt_available, reason="CVXOPT is not installed.")
def test_perform_phase_processing_gf(LP_solver="cvxopt"):
    """Perform LP phase processing on a single ray radar."""
    radar = pyart.testing.make_single_ray_radar()
    my_gatefilter = pyart.filters.GateFilter(radar)
    my_gatefilter.exclude_below("normalized_coherent_power", 0.5)
    my_gatefilter.exclude_below("cross_correlation_ratio", 0.8)
    phidp, kdp = pyart.correct.phase_proc_lp_gf(
        radar,
        gatefilter=my_gatefilter,
        LP_solver=LP_solver,
        doc=15,
        ncpts=20,
        system_phase=-140.1,
    )
    assert phidp["data"].max() < 360.0
    assert kdp["data"].max() < 6.0



"""
=====================================
Calculating and Plotting a Cloud Mask
=====================================

This example shows how to correct and plot reflectivity from an ARM
KAZR using a noise floor cloud mask.

"""
print(__doc__)

# Author: Adam Theisen and Zach Sherman
# License: BSD 3 clause

import matplotlib.pyplot as plt
import numpy as np
from open_radar_data import DATASETS

import pyart

############################
# **Read and plot raw data**
#
# First let's read and plot our dataset without any mask.

# Fetch and read in the ARM KAZR file.
filename = DATASETS.fetch("sgpkazrgeC1.a1.20190529.000002.cdf")
radar = pyart.aux_io.read_kazr(filename)

# Let's now take a look at reflectivity data prior to any corrections.
display = pyart.graph.RadarDisplay(radar)
display.plot("reflectivity_copol")
display.set_limits(xlim=(0, 55))
plt.show()

#################################################
# **Calculate cloud mask and plot corrected data**
#
# Now lets apply a mask by using the calc_cloud_mask function
# that will use a noise floor calculation from range and more
# to calculate the mask.

# First lets correct the data by calculating the mask.
cloud_mask_radar = pyart.correct.calc_cloud_mask(radar, "reflectivity_copol", "range")

# In this new radar object we should now have a new cloud mask field.
print(cloud_mask_radar.fields["cloud_mask_2"])

# Next we'll create a copy of the reflectivity field so we are not
# overwriting the original data.
cloud_mask_radar.add_field_like(
    "reflectivity_copol",
    "reflectivity_cloud_mask",
    cloud_mask_radar.fields["reflectivity_copol"]["data"].copy(),
)

# Now let's apply the mask to the copied reflectivity data.
cloud_mask_radar.fields["reflectivity_cloud_mask"]["data"][
    cloud_mask_radar.fields["cloud_mask_2"]["data"] == 0
] = np.nan

# And now we can plot the masked reflectivity field.
display = pyart.graph.RadarDisplay(cloud_mask_radar)
display.plot("reflectivity_cloud_mask")
display.set_limits(xlim=(0, 55))
plt.show()



"""
Calculation of storm-relative velocity from a radar object. Code written by
Edward C. Wolff. Modifications for single-sweep files suggested by Leanne
Blind.

"""

import math

import numpy as np

from ..config import get_field_name


def storm_relative_velocity(
    radar, direction=None, speed=None, field=None, u=None, v=None
):
    """
    This function calculates storm-relative Doppler velocities.

    Parameters
    ----------
    radar: Radar
        Radar object used.
    direction: float or string
        Direction of the storm motion vector (where north equals 0 degrees).
        Accepts a float or a string with the abbreviation of a cardinal or
        ordinal/intercardinal direction (for example: N, SE, etc.). If both
        speed/direction and u/v are specified, speed/direction will be used.
    speed: string
        Speed of the storm motion vector.
        Units should be identical to those in the provided radar
        object. If both speed/direction and u/v are specified, speed/direction
        will be used.
    field: string, optional
        Velocity field to use for storm-relative calculation. A value of None
        will use the default field name as defined in the Py-ART configuration
        file.
    u: float, optional
        U-component of the storm motion
    v: float, optional
        V-component of the storm motion

    Returns
    -------
    sr_data : dict
        Field dictionary containing storm-relative Doppler velocities in the
        same units as original velocities and the specified storm speed.
        Array is stored under the 'data' key.

    """
    # Parse the field parameter
    if field is None:
        field = get_field_name("velocity")

    # Obtain velocity data and copy the array
    sr_data = radar.fields[field]["data"].copy()

    # Specify cardinal directions that can be interpreted
    direction_dict = {
        "N": 0,
        "NE": 45,
        "E": 90,
        "SE": 135,
        "S": 180,
        "SW": 225,
        "W": 270,
        "NW": 315,
    }

    # Set the direction of the storm motion vector
    # When speed and direction are specified
    if direction is not None and speed is not None:
        if isinstance(direction, int) or isinstance(direction, float):
            alpha = direction
        elif isinstance(direction, str):
            if direction in direction_dict.keys():
                alpha = direction_dict[direction]
            else:
                raise ValueError("Direction string must be cardinal/ordinal direction")
        else:
            raise ValueError("Direction must be an integer, float, or string")
    # When u and v are specified
    elif u is not None:
        if v is not None:
            speed = np.sqrt((u**2) + (v**2))
            direction = 90 - np.rad2deg(math.atan2(v / speed, u / speed))
            if direction < 0:
                direction = direction + 360
        else:
            raise ValueError("Must specify both u and v components")
    else:
        raise ValueError("Must specify either speed and direction or u and v")

    # Calculates the storm relative velocities
    # If the radar file contains only one sweep (e.g. some research radars)
    if len(radar.sweep_number["data"]) == 1:
        sweep = 0
        start, end = radar.get_start_end(sweep)
        angle_array = radar.get_azimuth(sweep=sweep)
        ray_array = np.arange(start, end, 1)
        for count, ray in enumerate(ray_array):
            correction = speed * np.cos(np.deg2rad(alpha - angle_array[count]))
            sr_data[ray] = radar.fields[field]["data"][ray] - correction
    # If the radar file contains several sweeps, one volume scan (e.g. NEXRAD)
    else:
        for sweep in radar.sweep_number["data"]:
            start, end = radar.get_start_end(sweep)
            angle_array = radar.get_azimuth(sweep=sweep)
            ray_array = np.arange(start, end + 1, 1)
            for count, ray in enumerate(ray_array):
                correction = speed * np.cos(np.deg2rad(alpha - angle_array[count]))
                sr_data[ray] = radar.fields[field]["data"][ray] - correction

    return sr_data



"""
Classification of Precipitation Echoes in Radar Data.

Created on Thu Oct 12 23:12:19 2017
@author: Bhupendra Raut
@modifed: 11/19/2023
@references: 10.1109/TGRS.2020.2965649

.. autosummary::
    wavelet_reclass
    label_classes
    calc_scale_break
    atwt2d
"""

import numpy as np


def wavelet_reclass(
    grid,
    refl_field,
    level,
    zr_a,
    zr_b,
    core_wt_threshold,
    conv_wt_threshold,
    scale_break,
    min_reflectivity,
    conv_min_refl,
    conv_core_threshold,
):
    """
    Compute ATWT described as Raut et al (2008) and classify radar echoes using scheme of Raut et al (2020).
    First, convert dBZ to rain rates using standard Z-R relationship or user given coefficients. This is to
    transform the normally distributed dBZ to gamma-like distribution, enhancing the structure of the field.

    Parameters
    ----------
    dbz_data : ndarray
        2D array containing radar data. Last dimension should be levels.
    res_km : float
        Resolution of the radar data in km
    scale_break : int
        Calculated scale break between convective and stratiform scales. Dyadically spaced in grid pixels.

    Returns
    -------
    wt_class : ndarray
        Precipitation type classification: 0. N/A 1. stratiform/non-convective,
        2. convective cores and 3. moderate+transitional (mix) convective
        regions.
    """

    # Extract grid data, save mask and get the resolution in km
    try:
        dbz_data = grid.fields[refl_field]["data"][level, :, :]
    except:
        dbz_data = grid.fields[refl_field]["data"][:, :]

    # save the radar original mask for missing data.
    radar_mask = np.ma.getmask(dbz_data)

    wt_sum = conv_wavelet_sum(dbz_data, zr_a, zr_b, scale_break)

    wt_class = label_classes(
        wt_sum,
        dbz_data,
        core_wt_threshold,
        conv_wt_threshold,
        min_reflectivity,
        conv_min_refl,
        conv_core_threshold,
    )

    wt_class_ma = np.ma.masked_where(radar_mask, wt_class)  # add mask back
    wt_class_ma = wt_class_ma.squeeze()

    return wt_class_ma


def conv_wavelet_sum(dbz_data, zr_a, zr_b, scale_break):
    """
    Computes the sum of wavelet transform components for convective scales from dBZ data.

    Parameters
    ------------
    dbz_data : ndarray
        2D array containing radar dBZ data.
    zr_a, zr_b : float
        Coefficients for the Z-R relationship.
    res_km : float
        Resolution of the radar data in km.
    scale_break : int
        Calculated scale break (in pixels) between convective and stratiform scales

    Returns
    ---------
    wt_sum : ndarray
        Sum of convective scale wavelet transform components.
    """
    try:
        dbz_data = dbz_data.filled(0)
    except Exception:
        pass

    dbz_data[np.isnan(dbz_data)] = 0
    rr_data = ((10.0 ** (dbz_data / 10.0)) / zr_a) ** (1.0 / zr_b)

    wt, _ = atwt2d(rr_data, max_scale=scale_break)
    wt_sum = np.sum(wt, axis=(0))

    return wt_sum


def label_classes(
    wt_sum,
    dbz_data,
    core_wt_threshold,
    conv_wt_threshold,
    min_reflectivity,
    conv_min_refl,
    conv_core_threshold,
):
    """
    Labels classes using given thresholds:
        - 0: No precipitation or unclassified
        - 1: Stratiform/non-convective regions
        - 2: Transitional and mixed convective regions
        - 3: Convective cores

    Following hard coded values are optimized and validated using C-band radars
    over Darwin, Australia (2.5 km grid spacing) and tested for Solapur, India (1km grid spacing) [Raut et al. 2020].
    core_wt_threshold = 5  # WT value more than this is strong convection
    conv_wt_threshold = 2  # WT value for moderate convection
    min_reflectivity = 10  # pixels below this value are not classified.
     conv_min_refl = 30  # pixel below this value are not convective. This works for most cases.

    Parameters
    -----------
    wt_sum : ndarray
        Integrated wavelet transform
    vol_data : ndarray
        Array, vector or matrix of data

    Returns
    ---------
    wt_class : ndarray
        Precipitation type classification.
    """

    # I first used negative numbers to annotate the categories. Then multiply it by -1.
    wt_class = np.where(
        (wt_sum >= conv_wt_threshold) & (dbz_data >= conv_core_threshold), -3, 0
    )
    wt_class = np.where(
        (wt_sum >= core_wt_threshold) & (dbz_data >= conv_min_refl), -3, 0
    )
    wt_class = np.where(
        (wt_sum < core_wt_threshold)
        & (wt_sum >= conv_wt_threshold)
        & (dbz_data >= conv_min_refl),
        -2,
        wt_class,
    )
    wt_class = np.where((wt_class == 0) & (dbz_data >= min_reflectivity), -1, wt_class)

    wt_class = -1 * wt_class
    wt_class = np.where((wt_class == 0), np.nan, wt_class)

    return wt_class.astype(np.int32)


def calc_scale_break(res_meters, conv_scale_km):
    """
    Compute scale break for convection and stratiform regions. WT will be
    computed upto this scale and features will be designated as convection.

    Parameters
    -----------
    res_meters : float
        resolution of the image.
    conv_scale_km : float
        expected size of spatial variations due to convection.

    Returns
    --------
    dyadic scale break : int
        integer scale break in dyadic scale.
    """
    res_km = res_meters / 1000
    scale_break = np.log(conv_scale_km / res_km) / np.log(2) + 1

    return int(round(scale_break))


def atwt2d(data2d, max_scale=-1):
    """
    Computes a trous wavelet transform (ATWT). Computes ATWT of the 2D array
    up to max_scale. If max_scale is outside the boundaries, number of scales
    will be reduced.

    Data is mirrored at the boundaries. 'Negative WT are removed. Not tested
    for non-square data.

    @authors: Bhupendra A. Raut and Dileep M. Puranik
    @references: Press et al. (1992) Numerical Recipes in C.

    Parameters
    -----------
    data2d : ndarray
        2D image as array or matrix.
    max_scale :
        Computes wavelets up to max_scale. Leave blank for maximum possible
        scales.

    Returns
    ---------
    tuple of ndarray
        ATWT of input image and the final smoothed image or background image.
    """

    if not isinstance(data2d, np.ndarray):
        raise TypeError("The input data2d must be a numpy array.")

    data2d = data2d.squeeze()

    dims = data2d.shape
    min_dims = np.min(dims)
    max_possible_scales = int(np.floor(np.log(min_dims) / np.log(2)))

    if max_scale < 0 or max_possible_scales <= max_scale:
        max_scale = max_possible_scales - 1

    ny = dims[0]
    nx = dims[1]

    # For saving wt components
    wt = np.zeros((max_scale, ny, nx))

    temp1 = np.zeros(dims)
    temp2 = np.zeros(dims)

    sf = (0.0625, 0.25, 0.375)  # scaling function

    # start wavelet loop
    for scale in range(1, max_scale + 1):
        # print(scale)
        x1 = 2 ** (scale - 1)
        x2 = 2 * x1

        # Row-wise smoothing
        for i in range(0, nx):
            # find the indices for prev and next points on the line
            prev2 = abs(i - x2)
            prev1 = abs(i - x1)
            next1 = i + x1
            next2 = i + x2

            # If these indices are outside the image, "mirror" them
            # Sometime this causes issues at higher scales.
            if next1 > nx - 1:
                next1 = 2 * (nx - 1) - next1

            if next2 > nx - 1:
                next2 = 2 * (nx - 1) - next2

            if prev1 < 0 or prev2 < 0:
                prev1 = next1
                prev2 = next2

            for j in range(0, ny):
                left2 = data2d[j, prev2]
                left1 = data2d[j, prev1]
                right1 = data2d[j, next1]
                right2 = data2d[j, next2]
                temp1[j, i] = (
                    sf[0] * (left2 + right2)
                    + sf[1] * (left1 + right1)
                    + sf[2] * data2d[j, i]
                )

        # Column-wise smoothing
        for i in range(0, ny):
            prev2 = abs(i - x2)
            prev1 = abs(i - x1)
            next1 = i + x1
            next2 = i + x2

            # If these indices are outside the image use next values
            if next1 > ny - 1:
                next1 = 2 * (ny - 1) - next1
            if next2 > ny - 1:
                next2 = 2 * (ny - 1) - next2
            if prev1 < 0 or prev2 < 0:
                prev1 = next1
                prev2 = next2

            for j in range(0, nx):
                top2 = temp1[prev2, j]
                top1 = temp1[prev1, j]
                bottom1 = temp1[next1, j]
                bottom2 = temp1[next2, j]
                temp2[i, j] = (
                    sf[0] * (top2 + bottom2)
                    + sf[1] * (top1 + bottom1)
                    + sf[2] * temp1[i, j]
                )

        wt[scale - 1, :, :] = data2d - temp2
        data2d[:] = temp2

    return wt, data2d



"""
Plot Max-CAPPI

This module provides a function to plot a Maximum Constant Altitude Plan Position Indicator (Max-CAPPI)
from radar data using an xarray dataset. The function includes options for adding map features, range rings,
color bars, and customized visual settings.

Author: Syed Hamid Ali (@syedhamidali)
"""

__all__ = ["plot_maxcappi"]

import os
import warnings

import cartopy.crs as ccrs
import cartopy.feature as feat
import matplotlib.pyplot as plt
import numpy as np
from cartopy.mpl.gridliner import LATITUDE_FORMATTER, LONGITUDE_FORMATTER
from matplotlib.ticker import NullFormatter

warnings.filterwarnings("ignore")


def plot_maxcappi(
    grid,
    field,
    cmap=None,
    vmin=None,
    vmax=None,
    title=None,
    lat_lines=None,
    lon_lines=None,
    add_map=True,
    projection=None,
    colorbar=True,
    range_rings=False,
    dpi=100,
    savedir=None,
    show_figure=True,
    add_slogan=False,
    **kwargs,
):
    """
    Plot a Maximum Constant Altitude Plan Position Indicator (Max-CAPPI) using an xarray Dataset.

    Parameters
    ----------
    grid : pyart.core.Grid
        The grid object containing the radar data to be plotted.
    field : str
        The radar field to be plotted (e.g., "REF", "VEL", "WIDTH").
    cmap : str or matplotlib colormap, optional
        Colormap to use for the plot. Default is "HomeyerRainbow".
    vmin : float, optional
        Minimum value for the color scaling. Default is set to the minimum value of the data if not provided.
    vmax : float, optional
        Maximum value for the color scaling. Default is set to the maximum value of the data if not provided.
    title : str, optional
        Title of the plot. If None, the title is set to "Max-{field}".
    lat_lines : array-like, optional
        Latitude lines to be included in the plot. Default is calculated based on dataset coordinates.
    lon_lines : array-like, optional
        Longitude lines to be included in the plot. Default is calculated based on dataset coordinates.
    add_map : bool, optional
        Whether to include a map background in the plot. Default is True.
    projection : cartopy.crs.Projection, optional
        The map projection for the plot. Default is automatically determined based on dataset coordinates.
    colorbar : bool, optional
        Whether to include a colorbar in the plot. Default is True.
    range_rings : bool, optional
        Whether to include range rings at 50 km intervals. Default is False.
    dpi : int, optional
        DPI (dots per inch) for the plot. Default is 100.
    savedir : str, optional
        Directory where the plot will be saved. If None, the plot is not saved.
    show_figure : bool, optional
        Whether to display the plot. Default is True.
    add_slogan : bool, optional
        Whether to add a slogan like "Powered by Py-ART" to the plot. Default is False.
    **kwargs : dict, optional
        Additional keyword arguments to pass to matplotlib's `pcolormesh` function.

    Returns
    -------
    None
        This function does not return any value. It generates and optionally displays or saves a plot.

    Notes
    -----
    - The function extracts the maximum value across the altitude (z) dimension to create the Max-CAPPI.
    - It supports customizations such as map projections, color scales, and range rings.
    - If the radar_name attribute in the dataset is a byte string, it will be decoded and limited to 4 characters.
    - If add_map is True, map features and latitude/longitude lines are included.
    - The plot can be saved to a specified directory in PNG format.

    Author: Syed Hamid Ali (@syedhamidali)
    """

    ds = grid.to_xarray().squeeze()

    if lon_lines is None:
        lon_lines = np.arange(int(ds.lon.min().values), int(ds.lon.max().values) + 1)
    if lat_lines is None:
        lat_lines = np.arange(int(ds.lat.min().values), int(ds.lat.max().values) + 1)

    plt.rcParams.copy()
    plt.rcParams.update(
        {
            "font.weight": "bold",
            "axes.labelweight": "bold",
            "xtick.direction": "in",
            "ytick.direction": "in",
            "xtick.major.size": 10,
            "ytick.major.size": 10,
            "xtick.minor.size": 7,
            "ytick.minor.size": 7,
            "font.size": 14,
            "axes.linewidth": 2,
            "ytick.labelsize": 12,
            "xtick.labelsize": 12,
        }
    )

    max_c = ds[field].max(dim="z")
    max_x = ds[field].max(dim="y")
    max_y = ds[field].max(dim="x").T

    trgx = ds["x"].values
    trgy = ds["y"].values
    trgz = ds["z"].values

    max_height = int(np.floor(trgz.max()) / 1e3)
    sideticks = np.arange(max_height / 4, max_height + 1, max_height / 4).astype(int)

    if cmap is None:
        cmap = "HomeyerRainbow"
    if vmin is None:
        vmin = grid.fields[field]["data"].min()
    if vmax is None:
        vmax = grid.fields[field]["data"].max()
    if title is None:
        title = f"Max-{field.upper()[:3]}"

    def plot_range_rings(ax_xy, max_range):
        """
        Plots range rings at 50 km intervals.

        Parameters
        ----------
        ax_xy : matplotlib.axes.Axes
            The axis on which to plot the range rings.
        max_range : float
            The maximum range for the range rings.

        Returns
        -------
        None
        """
        background_color = ax_xy.get_facecolor()
        color = "k" if sum(background_color[:3]) / 3 > 0.5 else "w"

        for i, r in enumerate(np.arange(5e4, np.floor(max_range) + 1, 5e4)):
            label = f"Ring Dist. {int(r/1e3)} km" if i == 0 else None
            ax_xy.plot(
                r * np.cos(np.arange(0, 360) * np.pi / 180),
                r * np.sin(np.arange(0, 360) * np.pi / 180),
                color=color,
                ls="--",
                linewidth=0.4,
                alpha=0.3,
                label=label,
            )

        ax_xy.legend(loc="upper right", prop={"weight": "normal", "size": 8})

    def _get_projection(ds):
        """
        Determine the central latitude and longitude from a dataset
        and return the corresponding projection.

        Parameters
        ----------
        ds : xarray.Dataset
            The dataset from which to extract latitude and longitude
            information.

        Returns
        -------
        projection : cartopy.crs.Projection
            A Cartopy projection object centered on the extracted or
            calculated latitude and longitude.
        """

        def get_coord_or_attr(ds, coord_name, attr_name):
            """Helper function to get a coordinate or attribute, or
            calculate median if available.
            """
            if coord_name in ds:
                return (
                    ds[coord_name].values.item()
                    if ds[coord_name].values.ndim == 0
                    else ds[coord_name].values[0]
                )
            if f"origin_{coord_name}" in ds.coords:
                return ds.coords[f"origin_{coord_name}"].median().item()
            if f"radar_{coord_name}" in ds.coords:
                return ds.coords[f"radar_{coord_name}"].median().item()
            return ds.attrs.get(attr_name, None)

        lat_0 = get_coord_or_attr(
            ds, "latitude", "origin_latitude"
        ) or get_coord_or_attr(ds, "radar_latitude", "origin_latitude")
        lon_0 = get_coord_or_attr(
            ds, "longitude", "origin_longitude"
        ) or get_coord_or_attr(ds, "radar_longitude", "origin_longitude")

        if lat_0 is None or lon_0 is None:
            lat_0 = ds.lat.mean().item()
            lon_0 = ds.lon.mean().item()

        projection = ccrs.LambertAzimuthalEqualArea(lon_0, lat_0)
        return projection

    projection = _get_projection(ds)

    # FIG
    fig = plt.figure(figsize=[10.3, 10])
    left, bottom, width, height = 0.1, 0.1, 0.6, 0.2
    ax_xy = plt.axes((left, bottom, width, width), projection=projection)
    ax_x = plt.axes((left, bottom + width, width, height))
    ax_y = plt.axes((left + width, bottom, height, width))
    ax_cnr = plt.axes((left + width, bottom + width, left + left, height))
    if colorbar:
        ax_cb = plt.axes((left - 0.015 + width + height + 0.02, bottom, 0.02, width))

    # Set axis label formatters
    ax_x.xaxis.set_major_formatter(NullFormatter())
    ax_y.yaxis.set_major_formatter(NullFormatter())
    ax_cnr.yaxis.set_major_formatter(NullFormatter())
    ax_cnr.xaxis.set_major_formatter(NullFormatter())
    ax_x.set_ylabel("Height (km)", size=13)
    ax_y.set_xlabel("Height (km)", size=13)

    # Draw CAPPI
    plt.sca(ax_xy)
    xy = ax_xy.pcolormesh(trgx, trgy, max_c, cmap=cmap, vmin=vmin, vmax=vmax, **kwargs)

    # Add map features
    if add_map:
        map_features(ax_xy, lat_lines, lon_lines)

    ax_xy.minorticks_on()

    if range_rings:
        plot_range_rings(ax_xy, trgx.max())

    ax_xy.set_xlim(trgx.min(), trgx.max())
    ax_xy.set_ylim(trgx.min(), trgx.max())

    # Draw colorbar
    if colorbar:
        cb = plt.colorbar(xy, cax=ax_cb)
        cb.set_label(ds[field].attrs["units"], size=15)

    background_color = ax_xy.get_facecolor()
    color = "k" if sum(background_color[:3]) / 3 > 0.5 else "w"

    plt.sca(ax_x)
    plt.pcolormesh(trgx / 1e3, trgz / 1e3, max_x, cmap=cmap, vmin=vmin, vmax=vmax)
    plt.yticks(sideticks)
    ax_x.set_xlim(trgx.min() / 1e3, trgx.max() / 1e3)
    ax_x.grid(axis="y", lw=0.5, color=color, alpha=0.5, ls=":")
    ax_x.minorticks_on()

    plt.sca(ax_y)
    plt.pcolormesh(trgz / 1e3, trgy / 1e3, max_y, cmap=cmap, vmin=vmin, vmax=vmax)
    ax_y.set_xticks(sideticks)
    ax_y.set_ylim(trgx.min() / 1e3, trgx.max() / 1e3)
    ax_y.grid(axis="x", lw=0.5, color=color, alpha=0.5, ls=":")
    ax_y.minorticks_on()

    plt.sca(ax_cnr)
    plt.tick_params(
        axis="both",  # changes apply to both axes
        which="both",  # both major and minor ticks are affected
        bottom=False,  # ticks along the bottom edge are off
        top=False,  # ticks along the top edge are off
        left=False,
        right=False,
        labelbottom=False,
    )

    # Initialize an empty list to store the processed radar names
    full_title = []

    # Check if radar_name is a list (or list-like) or a simple string
    if isinstance(ds.attrs["radar_name"], list):
        # Iterate over each radar name in the list
        for name in ds.attrs["radar_name"]:
            # Decode if it's a byte string and take the first 4 characters
            if isinstance(name, bytes):
                site_title = name.decode("utf-8")[:4]
            else:
                site_title = name[:4]
            full_title.append(site_title)
    else:
        # Handle the case where radar_name is a single string
        site_title = ds.attrs["radar_name"][:4]
        full_title.append(site_title)

    # Join the processed radar names into a single string with commas separating them
    formatted_title = ", ".join(full_title)

    # Center-align text in the corner box
    plt.text(
        0.5,
        0.90,
        f"{formatted_title}",
        size=13,
        weight="bold",
        ha="center",
        va="center",
    )
    plt.text(0.5, 0.76, title, size=13, weight="bold", ha="center", va="center")
    plt.text(
        0.5,
        0.63,
        f"Max Range: {np.floor(trgx.max() / 1e3)} km",
        size=11,
        ha="center",
        va="center",
    )
    plt.text(
        0.5,
        0.47,
        f"Max Height: {np.floor(trgz.max() / 1e3)} km",
        size=11,
        ha="center",
        va="center",
    )
    plt.text(
        0.5,
        0.28,
        ds["time"].dt.strftime("%H:%M:%S Z").values.item(),
        weight="bold",
        size=16,
        ha="center",
        va="center",
    )
    plt.text(
        0.5,
        0.13,
        ds["time"].dt.strftime("%d %b, %Y UTC").values.item(),
        size=13.5,
        ha="center",
        va="center",
    )
    ax_xy.set_aspect("auto")

    if add_slogan:
        fig.text(
            0.1,
            0.06,
            "Powered by Py-ART",  # Coordinates close to (0, 0) for lower-left corner
            fontsize=9,
            fontname="Courier New",
            # bbox=dict(facecolor='none', boxstyle='round,pad=0.5')
        )

    if savedir is not None:
        radar_name = ds.attrs.get("instrument_name", "Radar")
        time_str = ds["time"].dt.strftime("%Y%m%d%H%M%S").values.item()
        figname = f"{savedir}{os.sep}{title}_{radar_name}_{time_str}.png"
        plt.savefig(fname=figname, dpi=dpi, bbox_inches="tight")
        print(f"Figure(s) saved as {figname}")

    # plt.rcParams.update(original_rc_params)
    plt.rcdefaults()

    if show_figure:
        plt.show()
    else:
        plt.close()


def map_features(ax, lat_lines, lon_lines):
    """
    Adds map features and gridlines to the plot.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axis on which to add map features and gridlines.
    lat_lines : array-like
        Latitude lines for the gridlines.
    lon_lines : array-like
        Longitude lines for the gridlines.

    Returns
    -------
    None
    """
    background_color = ax.get_facecolor()
    color = "k" if sum(background_color[:3]) / 3 > 0.5 else "w"

    # Labeling gridlines depending on the projection
    if isinstance(ax.projection, (ccrs.PlateCarree, ccrs.Mercator)):
        gl = ax.gridlines(
            xlocs=lon_lines,
            ylocs=lat_lines,
            linewidth=1,
            alpha=0.5,
            linestyle="--",
            draw_labels=True,
        )
        gl.top_labels = False
        gl.right_labels = False
        gl.xlines = False
        gl.ylines = False
        gl.xlabel_style = {"color": color}
        gl.ylabel_style = {"color": color}
        ax.add_feature(feat.COASTLINE, alpha=0.8, lw=1, ec=color)
        ax.add_feature(feat.BORDERS, alpha=0.7, lw=0.7, ls="--", ec=color)
        ax.add_feature(
            feat.STATES.with_scale("10m"), alpha=0.6, lw=0.5, ls=":", ec=color
        )

    elif isinstance(
        ax.projection, (ccrs.LambertConformal, ccrs.LambertAzimuthalEqualArea)
    ):
        ax.figure.canvas.draw()
        gl = ax.gridlines(
            crs=ccrs.PlateCarree(),
            xlocs=lon_lines,
            ylocs=lat_lines,
            linewidth=1,
            alpha=0.5,
            linestyle="--",
            draw_labels=False,
        )
        gl.xlines = False
        gl.ylines = False
        gl.xlabel_style = {"color": color}
        gl.ylabel_style = {"color": color}
        ax.add_feature(feat.COASTLINE, alpha=0.8, lw=1, ec=color)
        ax.add_feature(feat.BORDERS, alpha=0.7, lw=0.7, ls="--", ec=color)
        ax.add_feature(
            feat.STATES.with_scale("10m"), alpha=0.6, lw=0.5, ls=":", ec=color
        )
        # Label the end-points of the gridlines using custom tick makers
        ax.xaxis.set_major_formatter(LONGITUDE_FORMATTER)
        ax.yaxis.set_major_formatter(LATITUDE_FORMATTER)
        from pyart.graph.gridmapdisplay import lambert_xticks, lambert_yticks

        lambert_xticks(ax, lon_lines)
        lambert_yticks(ax, lat_lines)
    else:
        ax.gridlines(xlocs=lon_lines, ylocs=lat_lines)



""" Unit Tests for Py-ART's io/mdv_radar.py module. """

import numpy as np
import pytest
from numpy.ma.core import MaskedArray
from numpy.testing import assert_almost_equal

import pyart

############################################
# read_mdv tests (verify radar attributes) #
############################################

# read in the sample file and create a a Radar object
radar = pyart.io.read_mdv(pyart.testing.MDV_PPI_FILE)


# time attribute
def test_time():
    assert "comment" in radar.time.keys()
    assert "long_name" in radar.time.keys()
    assert "standard_name" in radar.time.keys()
    assert "units" in radar.time.keys()
    assert "calendar" in radar.time.keys()
    assert "data" in radar.time.keys()
    assert radar.time["units"] == "seconds since 2011-05-20T11:01:00Z"
    assert radar.time["data"].shape == (360,)
    assert_almost_equal(radar.time["data"][200], 187, 0)


# range attribute
def test_range():
    assert "long_name" in radar.range
    assert "standard_name" in radar.range
    assert "meters_to_center_of_first_gate" in radar.range
    assert "meters_between_gates" in radar.range
    assert "units" in radar.range
    assert "data" in radar.range
    assert "spacing_is_constant" in radar.range
    assert radar.range["data"].shape == (110,)
    assert_almost_equal(radar.range["data"][0], 118, 0)


# fields attribute is tested later


# metadata attribute
def test_metadata():
    assert "instrument_name" in radar.metadata
    assert "source" in radar.metadata


# scan_type attribute
def test_scan_type():
    assert radar.scan_type == "ppi"


# latitude attribute
def test_latitude():
    assert "data" in radar.latitude
    assert "standard_name" in radar.latitude
    assert "units" in radar.latitude
    assert radar.latitude["data"].shape == (1,)
    assert_almost_equal(radar.latitude["data"], 37, 0)


# longitude attribute
def test_longitude():
    assert "data" in radar.longitude
    assert "standard_name" in radar.longitude
    assert "units" in radar.longitude
    assert radar.longitude["data"].shape == (1,)
    assert_almost_equal(radar.longitude["data"], -97, 0)


# altitude attribute
def test_altitude():
    assert "data" in radar.altitude
    assert "standard_name" in radar.altitude
    assert "units" in radar.altitude
    assert "positive" in radar.altitude
    assert radar.altitude["data"].shape == (1,)
    assert_almost_equal(radar.altitude["data"], 328, 0)


# altitude_agl attribute
def test_altitude_agl():
    assert radar.altitude_agl is None


# sweep_number attribute
def test_sweep_number():
    assert "standard_name" in radar.sweep_number
    assert np.all(radar.sweep_number["data"] == range(1))


# sweep_mode attribute
def test_sweep_mode():
    assert "standard_name" in radar.sweep_mode
    assert radar.sweep_mode["data"].shape == (1,)
    assert radar.sweep_mode["data"].dtype.char == "S"
    assert np.all(radar.sweep_mode["data"] == [b"azimuth_surveillance"])


# fixed_angle attribute
def test_fixed_angle():
    assert "standard_name" in radar.fixed_angle
    assert "units" in radar.fixed_angle
    assert radar.fixed_angle["data"].shape == (1,)
    assert_almost_equal(radar.fixed_angle["data"][0], 0.75, 2)


# sweep_start_ray_index attribute
def test_sweep_start_ray_index():
    assert "long_name" in radar.sweep_start_ray_index
    assert radar.sweep_start_ray_index["data"].shape == (1,)
    assert_almost_equal(radar.sweep_start_ray_index["data"][0], 0, 0)


# sweep_end_ray_index attribute
def test_sweep_end_ray_index():
    assert "long_name" in radar.sweep_end_ray_index
    assert radar.sweep_end_ray_index["data"].shape == (1,)
    assert_almost_equal(radar.sweep_end_ray_index["data"][0], 359, 0)


# target_scan_rate attribute
def test_target_scan_rate():
    assert radar.target_scan_rate is None


# azimuth attribute
def test_azimuth():
    assert "standard_name" in radar.azimuth
    assert "long_name" in radar.azimuth
    assert "units" in radar.azimuth
    assert "axis" in radar.azimuth
    assert_almost_equal(radar.azimuth["data"][0], 0, 0)
    assert_almost_equal(radar.azimuth["data"][10], 10.0, 0)


# elevation attribute
def test_elevation():
    assert "standard_name" in radar.elevation
    assert "long_name" in radar.azimuth
    assert "units" in radar.elevation
    assert "axis" in radar.elevation
    assert radar.elevation["data"].shape == (360,)
    assert_almost_equal(radar.elevation["data"][0], 0.75, 2)


# scan_rate attribute
def test_scan_rate():
    assert radar.scan_rate is None


# antenna_transition attribute
def test_antenna_transition():
    assert radar.antenna_transition is None


# instrument_parameters attribute
@pytest.mark.parametrize(
    "keys", ["prt", "unambiguous_range", "prt_mode", "nyquist_velocity"]
)
def test_instument_parameters(keys):
    # instrument_parameter sub-convention
    description = f"instrument_parameters: {keys}"
    check_instrument_parameter.description = description
    check_instrument_parameter(keys)


def check_instrument_parameter(param):
    assert param in radar.instrument_parameters
    param_dic = radar.instrument_parameters[param]
    assert param_dic["meta_group"] == "instrument_parameters"


# radar_parameters attribute
@pytest.mark.parametrize("keys", ["radar_beam_width_h", "radar_beam_width_v"])
def test_radar_parameters(keys):
    # radar_parameter sub-convention
    description = f"radar_parameters: {keys}"
    check_radar_parameter.description = description
    check_radar_parameter(keys)


def check_radar_parameter(param):
    assert param in radar.instrument_parameters
    param_dic = radar.instrument_parameters[param]
    assert param_dic["meta_group"] == "radar_parameters"


# radar_calibration attribute
def test_radar_calibration():
    assert radar.radar_calibration is None


# ngates attribute
def test_ngates():
    assert radar.ngates == 110


# nrays attribute
def test_nrays():
    assert radar.nrays == 360


# nsweeps attribute
def test_nsweeps():
    assert radar.nsweeps == 1


####################
# fields attribute #
####################


@pytest.mark.parametrize(
    "field",
    [
        "reflectivity",
    ],
)
def test_field_dics(field):
    description = f"field : {field}, dictionary"
    check_field_dic.description = description
    check_field_dic(field)


def check_field_dic(field):
    """Check that the required keys are present in a field dictionary."""
    assert "standard_name" in radar.fields[field]
    assert "units" in radar.fields[field]
    assert "_FillValue" in radar.fields[field]
    assert "coordinates" in radar.fields[field]


@pytest.mark.parametrize(
    "field",
    [
        "reflectivity",
    ],
)
def test_field_shapes(field):
    description = f"field : {field}, shape"
    check_field_shape.description = description
    check_field_shape(field)


def check_field_shape(field):
    assert radar.fields[field]["data"].shape == (360, 110)


fields = {
    "reflectivity": MaskedArray,
}


@pytest.mark.parametrize("field, field_type", fields.items(), ids=list(fields.keys()))
def test_field_types(field, field_type):
    description = f"field : {field}, type"
    check_field_type.description = description
    check_field_type(field, field_type)


def check_field_type(field, field_type):
    assert type(radar.fields[field]["data"]) is field_type


fields = {"reflectivity": 24.0}


@pytest.mark.parametrize("field, field_value", fields.items(), ids=list(fields.keys()))
def test_field_first_points(field, field_value):
    # these values can be found using:
    # [round(radar.fields[f]['data'][0,0]) for f in radar.fields]
    description = f"field : {field}, first point"
    check_field_first_point.description = description
    check_field_first_point(field, field_value)


def check_field_first_point(field, value):
    assert_almost_equal(radar.fields[field]["data"][0, 0], value, 0)


#############
# RHI tests #
#############

RADAR_RHI = pyart.io.read_mdv(pyart.testing.MDV_RHI_FILE, delay_field_loading=True)


# nsweeps attribute
def test_rhi_nsweeps():
    assert RADAR_RHI.nsweeps == 1


# sweep_number attribute
def test_rhi_sweep_number():
    assert "standard_name" in RADAR_RHI.sweep_number
    assert np.all(RADAR_RHI.sweep_number["data"] == range(1))


# sweep_mode attribute
def test_rhi_sweep_mode():
    assert "standard_name" in RADAR_RHI.sweep_mode
    assert RADAR_RHI.sweep_mode["data"].shape == (1,)
    assert np.all(RADAR_RHI.sweep_mode["data"] == [b"rhi"])


# fixed_angle attribute
def test_rhi_fixed_angle():
    assert "standard_name" in RADAR_RHI.fixed_angle
    assert "units" in RADAR_RHI.fixed_angle
    assert RADAR_RHI.fixed_angle["data"].shape == (1,)
    assert_almost_equal(RADAR_RHI.fixed_angle["data"][0], 189.00, 2)


# sweep_start_ray_index attribute
def test_rhi_sweep_start_ray_index():
    assert "long_name" in RADAR_RHI.sweep_start_ray_index
    assert RADAR_RHI.sweep_start_ray_index["data"].shape == (1,)
    assert_almost_equal(RADAR_RHI.sweep_start_ray_index["data"][0], 0, 0)


# sweep_end_ray_index attribute
def test_rhi_sweep_end_ray_index():
    assert "long_name" in RADAR_RHI.sweep_end_ray_index
    assert RADAR_RHI.sweep_end_ray_index["data"].shape == (1,)
    assert_almost_equal(RADAR_RHI.sweep_end_ray_index["data"][0], 282, 0)


# azimuth attribute
def test_rhi_azimuth():
    assert "standard_name" in RADAR_RHI.azimuth
    assert "long_name" in RADAR_RHI.azimuth
    assert "units" in RADAR_RHI.azimuth
    assert "axis" in RADAR_RHI.azimuth
    assert_almost_equal(RADAR_RHI.azimuth["data"][0], 189, 0)
    assert_almost_equal(RADAR_RHI.azimuth["data"][10], 189, 0)


# elevation attribute
def test_rhi_elevation():
    assert "standard_name" in RADAR_RHI.elevation
    assert "long_name" in RADAR_RHI.azimuth
    assert "units" in RADAR_RHI.elevation
    assert "axis" in RADAR_RHI.elevation
    assert RADAR_RHI.elevation["data"].shape == (283,)
    assert_almost_equal(RADAR_RHI.elevation["data"][0], 19.6, 2)


# field data
def test_rhi_field_data():
    assert_almost_equal(RADAR_RHI.fields["reflectivity"]["data"][0, 0], 23.93, 2)


def test_open_from_file_obj():
    fh = open(pyart.testing.MDV_PPI_FILE, "rb")
    pyart.io.read_mdv(pyart.testing.MDV_PPI_FILE)
    fh.close()


def test_radar_exclude_fields():
    # skip fields
    radar = pyart.io.read_mdv(
        pyart.testing.MDV_PPI_FILE, exclude_fields=["reflectivity"]
    )
    assert "reflectivity" not in radar.fields



"""
Functions for echo classification.

"""

from copy import deepcopy
from warnings import warn

import numpy as np

# Local imports
# Local imports
from ..config import get_field_name, get_fillvalue, get_metadata
from ..core import Grid
from ..util.radar_utils import ma_broadcast_to
from ._echo_class import _feature_detection, steiner_class_buff
from ._echo_class_wt import calc_scale_break, wavelet_reclass


def steiner_conv_strat(
    grid,
    dx=None,
    dy=None,
    intense=42.0,
    work_level=3000.0,
    peak_relation="default",
    area_relation="medium",
    bkg_rad=11000.0,
    use_intense=True,
    fill_value=None,
    refl_field=None,
):
    """
    Partition reflectivity into convective-stratiform using the Steiner et
    al. (1995) algorithm.

    Parameters
    ----------
    grid : Grid
        Grid containing reflectivity field to partition.
    dx, dy : float, optional
        The x- and y-dimension resolutions in meters, respectively. If None
        the resolution is determined from the first two axes values.
    intense : float, optional
        The intensity value in dBZ. Grid points with a reflectivity
        value greater or equal to the intensity are automatically
        flagged as convective. See reference for more information.
    work_level : float, optional
        The working level (separation altitude) in meters. This is the height
        at which the partitioning will be done, and should minimize bright band
        contamination. See reference for more information.
    peak_relation : 'default' or 'sgp', optional
        The peakedness relation. See reference for more information.
    area_relation : 'small', 'medium', 'large', or 'sgp', optional
        The convective area relation. See reference for more information.
    bkg_rad : float, optional
        The background radius in meters. See reference for more information.
    use_intense : bool, optional
        True to use the intensity criteria.
    fill_value : float, optional
         Missing value used to signify bad data points. A value of None
         will use the default fill value as defined in the Py-ART
         configuration file.
    refl_field : str, optional
         Field in grid to use as the reflectivity during partitioning. None
         will use the default reflectivity field name from the Py-ART
         configuration file.

    Returns
    -------
    eclass : dict
        Steiner convective-stratiform classification dictionary.

    References
    ----------
    Steiner, M. R., R. A. Houze Jr., and S. E. Yuter, 1995: Climatological
    Characterization of Three-Dimensional Storm Structure from Operational
    Radar and Rain Gauge Data. J. Appl. Meteor., 34, 1978-2007.

    """
    # Get fill value
    if fill_value is None:
        fill_value = get_fillvalue()

    # Parse field parameters
    if refl_field is None:
        refl_field = get_field_name("reflectivity")

    # parse dx and dy
    if dx is None:
        dx = grid.x["data"][1] - grid.x["data"][0]
    if dy is None:
        dy = grid.y["data"][1] - grid.y["data"][0]

    # Get coordinates
    x = grid.x["data"]
    y = grid.y["data"]
    z = grid.z["data"]

    # Get reflectivity data
    ze = np.ma.copy(grid.fields[refl_field]["data"])
    ze = ze.filled(np.nan)

    eclass = steiner_class_buff(
        ze,
        x,
        y,
        z,
        dx=dx,
        dy=dy,
        bkg_rad=bkg_rad,
        work_level=work_level,
        intense=intense,
        peak_relation=peak_relation,
        area_relation=area_relation,
        use_intense=use_intense,
    )

    return {
        "data": eclass.astype(np.int32),
        "standard_name": "echo_classification",
        "long_name": "Steiner echo classification",
        "valid_min": 0,
        "valid_max": 2,
        "comment_1": (
            "Convective-stratiform echo "
            "classification based on "
            "Steiner et al. (1995)"
        ),
        "comment_2": ("0 = Undefined, 1 = Stratiform, " "2 = Convective"),
    }


def conv_strat_yuter(
    grid,
    dx=None,
    dy=None,
    level_m=None,
    always_core_thres=42,
    bkg_rad_km=11,
    use_cosine=True,
    max_diff=5,
    zero_diff_cos_val=55,
    scalar_diff=1.5,
    use_addition=True,
    calc_thres=0.75,
    weak_echo_thres=5.0,
    min_dBZ_used=5.0,
    dB_averaging=True,
    remove_small_objects=True,
    min_km2_size=10,
    val_for_max_conv_rad=30,
    max_conv_rad_km=5.0,
    cs_core=3,
    nosfcecho=0,
    weakecho=3,
    sf=1,
    conv=2,
    refl_field=None,
    estimate_flag=True,
    estimate_offset=5,
):
    """
    Partition reflectivity into convective-stratiform using the Yuter et al. (2005)
    and Yuter and Houze (1997) algorithm.

    Parameters
    ----------
    grid : Grid
        Grid containing reflectivity field to partition.
    dx, dy : float, optional
        The x- and y-dimension resolutions in meters, respectively. If None
        the resolution is determined from the first two axes values parsed from grid object.
    level_m : float, optional
        Desired height in meters to classify with convective stratiform algorithm.
    always_core_thres : float, optional
        Threshold for points that are always convective. All values above the threshold are classifed as convective
        See Yuter et al. (2005) for more detail.
    bkg_rad_km : float, optional
        Radius to compute background reflectivity in kilometers. Default is 11 km. Recommended to be at least 3 x
        grid spacing
    use_cosine : bool, optional
        Boolean used to determine if a cosine scheme (see Yuter and Houze (1997)) should be used for identifying
        convective cores (True) or if a simpler scalar scheme (False) should be used.
    max_diff : float, optional
        Maximum difference between background average and reflectivity in order to be classified as convective.
        "a" value in Eqn. B1 in Yuter and Houze (1997)
    zero_diff_cos_val : float, optional
        Value where difference between background average and reflectivity is zero in the cosine function
        "b" value in Eqn. B1 in Yuter and Houze (1997)
    scalar_diff : float, optional
        If using a scalar difference scheme, this value is the multiplier or addition to the background average
    use_addition : bool, optional
        Determines if a multiplier (False) or addition (True) in the scalar difference scheme should be used
    calc_thres : float, optional
        Minimum percentage of points needed to be considered in background average calculation
    weak_echo_thres : float, optional
        Threshold for determining weak echo. All values below this threshold will be considered weak echo
    min_dBZ_used : float, optional
        Minimum dBZ value used for classification. All values below this threshold will be considered no surface echo
        See Yuter and Houze (1997) and Yuter et al. (2005) for more detail.
    dB_averaging : bool, optional
        True if using dBZ reflectivity values that need to be converted to linear Z before averaging. False for
        other non-dBZ values (i.e. snow rate)
    remove_small_objects : bool, optional
        Determines if small objects should be removed from convective core array. Default is True.
    min_km2_size : float, optional
        Minimum size of convective cores to be considered. Cores less than this size will be removed. Default is 10
        km^2.
    val_for_max_conv_rad : float, optional
        dBZ for maximum convective radius. Convective cores with values above this will have the maximum convective
        radius
    max_conv_rad_km : float, optional
        Maximum radius around convective cores to classify as convective. Default is 5 km
    cs_core : int, optional
        Value for points classified as convective cores
    nosfcecho : int, optional
        Value for points classified as no surface echo, based on min_dBZ_used
    weakecho : int, optional
        Value for points classified as weak echo, based on weak_echo_thres
    sf : int, optional
        Value for points classified as stratiform
    conv : int, optional
        Value for points classified as convective
    refl_field : str, optional
        Field in grid to use as the reflectivity during partitioning. None will use the default reflectivity
        field name from the Py-ART configuration file.
    estimate_flag : bool, optional
        Determines if over/underestimation should be applied. If true, the algorithm will also be run on the same field
        wih the estimate_offset added and the same field with the estimate_offset subtracted.
        Default is True (recommended)
    estimate_offset : float, optional
        Value used to offset the reflectivity values by for the over/underestimation application. Default value is 5
        dBZ.

    Returns
    -------
    convsf_dict : dict
        Convective-stratiform classification dictionary.

    References
    ----------
    Yuter, S. E., and R. A. Houze, Jr., 1997: Measurements of raindrop size
    distributions over the Pacific warm pool and implications for Z-R relations.
    J. Appl. Meteor., 36, 847-867.
    https://doi.org/10.1175/1520-0450(1997)036%3C0847:MORSDO%3E2.0.CO;2

    Yuter, S. E., R. A. Houze, Jr., E. A. Smith, T. T. Wilheit, and E. Zipser,
    2005: Physical characterization of tropical oceanic convection observed in
    KWAJEX. J. Appl. Meteor., 44, 385-415. https://doi.org/10.1175/JAM2206.1

    """

    warn(
        "This function will be deprecated in Py-ART 2.0."
        " Please use feature_detection function.",
        DeprecationWarning,
    )

    feature_dict = feature_detection(
        grid,
        dx=dx,
        dy=dy,
        level_m=level_m,
        always_core_thres=always_core_thres,
        bkg_rad_km=bkg_rad_km,
        use_cosine=use_cosine,
        max_diff=max_diff,
        zero_diff_cos_val=zero_diff_cos_val,
        scalar_diff=scalar_diff,
        use_addition=use_addition,
        calc_thres=calc_thres,
        weak_echo_thres=weak_echo_thres,
        min_val_used=min_dBZ_used,
        dB_averaging=dB_averaging,
        remove_small_objects=remove_small_objects,
        min_km2_size=min_km2_size,
        binary_close=False,
        val_for_max_rad=val_for_max_conv_rad,
        max_rad_km=max_conv_rad_km,
        core_val=cs_core,
        nosfcecho=nosfcecho,
        weakecho=weakecho,
        bkgd_val=sf,
        feat_val=conv,
        field=refl_field,
        estimate_flag=estimate_flag,
        estimate_offset=5,
        overest_field=None,
        underest_field=None,
    )

    return feature_dict


def feature_detection(
    grid,
    dx=None,
    dy=None,
    level_m=None,
    always_core_thres=42,
    bkg_rad_km=11,
    use_cosine=True,
    max_diff=5,
    zero_diff_cos_val=55,
    scalar_diff=1.5,
    use_addition=True,
    calc_thres=0.75,
    weak_echo_thres=5.0,
    min_val_used=5.0,
    dB_averaging=True,
    remove_small_objects=True,
    min_km2_size=10,
    binary_close=False,
    val_for_max_rad=30,
    max_rad_km=5.0,
    core_val=3,
    nosfcecho=0,
    weakecho=3,
    bkgd_val=1,
    feat_val=2,
    field=None,
    estimate_flag=True,
    estimate_offset=5,
    overest_field=None,
    underest_field=None,
):
    """
    This function can be used to detect features in a field (e.g. reflectivity, rain rate, snow rate,
    etc.) described by Tomkins et al. (2024) and based on original convective-stratiform algorithms developed by
    Steiner et al. (1995), Yuter et al. (2005) and Yuter and Houze (1997) algorithm.

    Author: Laura Tomkins (lauramtomkins@gmail.com)

    Parameters
    ----------
    grid : Grid
        Grid containing reflectivity field to partition.
    dx, dy : float, optional
        The x- and y-dimension resolutions in meters, respectively. If None
        the resolution is determined from the first two axes values parsed from grid object.
    level_m : float, optional
        Desired height in meters to run feature detection algorithm.
    always_core_thres : float, optional
        Threshold for points that are always features. All values above the threshold are classified as features.
    bkg_rad_km : float, optional
        Radius to compute background reflectivity in kilometers. Default is 11 km. Recommended to be at least 3 x
        grid spacing
    use_cosine : bool, optional
        Boolean used to determine if a cosine scheme (see Yuter and Houze (1997)) should be used for identifying
        cores (True) or if a simpler scalar scheme (False) should be used.
    max_diff : float, optional
        Maximum difference between background average and reflectivity in order to be classified as features.
        "a" value in Eqn. B1 in Yuter and Houze (1997)
    zero_diff_cos_val : float, optional
        Value where difference between background average and reflectivity is zero in the cosine function
        "b" value in Eqn. B1 in Yuter and Houze (1997)
    scalar_diff : float, optional
        If using a scalar difference scheme, this value is the multiplier or addition to the background average
    use_addition : bool, optional
        Determines if a multiplier (False) or addition (True) in the scalar difference scheme should be used
    calc_thres : float, optional
        Minimum percentage of points needed to be considered in background average calculation
    weak_echo_thres : float, optional
        Threshold for determining weak echo. All values below this threshold will be considered weak echo
    min_val_used : float, optional
        Minimum value used for classification. All values below this threshold will be considered no surface echo
        See Yuter and Houze (1997) and Yuter et al. (2005) for more detail. Units based on input field
    dB_averaging : bool, optional
        True if using dBZ reflectivity values that need to be converted to linear Z before averaging. False for
        other non-dBZ values (i.e. snow rate)
    remove_small_objects : bool, optional
        Determines if small objects should be removed from core array. Default is True.
    min_km2_size : float, optional
        Minimum size of Cores to be considered. Cores less than this size will be removed. Default is 10 km^2.
    binary_close : bool, optional
        Determines if a binary closing should be performed on the cores. Default is False.
    val_for_max_rad : float, optional
        value used for maximum radius. Cores with values above this will have the maximum radius incorporated.
    max_rad_km : float, optional
        Maximum radius around cores to classify as feature. Default is 5 km
    core_val : int, optional
        Value for points classified as cores
    nosfcecho : int, optional
        Value for points classified as no surface echo, based on min_val_used
    weakecho : int, optional
        Value for points classified as weak echo, based on weak_echo_thres.
    bkgd_val : int, optional
        Value for points classified as background echo.
    feat_val : int, optional
        Value for points classified as features.
    field : str, optional
        Field in grid to find objects in. None will use the default reflectivity field name from the Py-ART
        configuration file.
    estimate_flag : bool, optional
        Determines if over/underestimation should be applied. If true, the algorithm will also be run on the same field
        wih the estimate_offset added and the same field with the estimate_offset subtracted.
        Default is True (recommended)
    estimate_offset : float, optional
        Value used to offset the field values by for the over/underestimation application. Default value is 5 dBZ.
    overest_field : str, optional
        Name of field in grid object used to calculate the overestimate if estimate_flag is True.
    underest_field : str, optional
        Name of field in grid object used to calculate the underestimate if estimate_flag is True.

    Returns
    -------
    feature_dict : dict
        Feature detection classification dictionary.

    References
    ----------
    Steiner, M. R., R. A. Houze Jr., and S. E. Yuter, 1995: Climatological
    Characterization of Three-Dimensional Storm Structure from Operational
    Radar and Rain Gauge Data. J. Appl. Meteor., 34, 1978-2007.

    Yuter, S. E., and R. A. Houze, Jr., 1997: Measurements of raindrop size
    distributions over the Pacific warm pool and implications for Z-R relations.
    J. Appl. Meteor., 36, 847-867.
    https://doi.org/10.1175/1520-0450(1997)036%3C0847:MORSDO%3E2.0.CO;2

    Yuter, S. E., R. A. Houze, Jr., E. A. Smith, T. T. Wilheit, and E. Zipser,
    2005: Physical characterization of tropical oceanic convection observed in
    KWAJEX. J. Appl. Meteor., 44, 385-415. https://doi.org/10.1175/JAM2206.1

    Tomkins, L. M., Yuter, S. E., and Miller, M. A., 2024: Dual adaptive differential
    threshold method for automated detection of faint and strong echo features
    in radar observations of winter storms. Atmos. Meas. Tech., 17, 3377–3399.
    https://doi.org/10.5194/amt-17-3377-2024

    """

    # Maxmimum radius must be less than 5 km
    if max_rad_km > 5:
        print("Max radius must be less than 5 km, exiting")
        raise

    # Parse field parameters
    if field is None:
        field = get_field_name("reflectivity")
        dB_averaging = True

    # parse dx and dy if None
    if dx is None:
        dx = grid.x["data"][1] - grid.x["data"][0]
    if dy is None:
        dy = grid.y["data"][1] - grid.y["data"][0]

    # add catch for background radius size
    if bkg_rad_km * 1000 < 2 * dx or bkg_rad_km * 1000 < 2 * dy:
        print(
            "Background radius for averaging must be at least 2 times dx and dy, exiting"
        )
        raise

    # Get coordinates
    z = grid.z["data"]

    # Get reflectivity data at desired level
    if level_m is None:
        try:
            ze = np.ma.copy(grid.fields[field]["data"][0, :, :])
        except:
            ze = np.ma.copy(grid.fields[field]["data"][:, :])
    else:
        zslice = np.argmin(np.abs(z - level_m))
        ze = np.ma.copy(grid.fields[field]["data"][zslice, :, :])

    # run feature detection algorithm
    _, _, feature_best = _feature_detection(
        ze,
        dx,
        dy,
        always_core_thres=always_core_thres,
        bkg_rad_km=bkg_rad_km,
        use_cosine=use_cosine,
        max_diff=max_diff,
        zero_diff_cos_val=zero_diff_cos_val,
        scalar_diff=scalar_diff,
        use_addition=use_addition,
        calc_thres=calc_thres,
        weak_echo_thres=weak_echo_thres,
        min_val_used=min_val_used,
        dB_averaging=dB_averaging,
        remove_small_objects=remove_small_objects,
        min_km2_size=min_km2_size,
        binary_close=binary_close,
        val_for_max_rad=val_for_max_rad,
        max_rad_km=max_rad_km,
        core_val=core_val,
        nosfcecho=nosfcecho,
        weakecho=weakecho,
        bkgd_val=bkgd_val,
        feat_val=feat_val,
    )

    # put data into a dictionary to be added as a field
    feature_dict = {
        "feature_detection": {
            "data": feature_best[None, :, :],
            "standard_name": "feature_detection",
            "long_name": "Feature Detection",
            "valid_min": 0,
            "valid_max": 3,
            "comment_1": (
                f"{nosfcecho} = No surface echo/Undefined, {bkgd_val} = Background echo, {feat_val} = Features, {weakecho} = weak echo"
            ),
        }
    }

    # If estimation is True, run the algorithm on the field with offset subtracted and the field with the offset added
    if estimate_flag:
        # get underestimate field
        if underest_field is None:
            under_field = ze - estimate_offset
        elif underest_field is not None:
            under_field = np.ma.copy(grid.fields[underest_field]["data"][0, :, :])

        # run algorithm to get feature detection underestimate
        _, _, feature_under = _feature_detection(
            under_field,
            dx,
            dy,
            always_core_thres=always_core_thres,
            bkg_rad_km=bkg_rad_km,
            use_cosine=use_cosine,
            max_diff=max_diff,
            zero_diff_cos_val=zero_diff_cos_val,
            scalar_diff=scalar_diff,
            use_addition=use_addition,
            calc_thres=calc_thres,
            weak_echo_thres=weak_echo_thres,
            min_val_used=min_val_used,
            dB_averaging=dB_averaging,
            remove_small_objects=remove_small_objects,
            min_km2_size=min_km2_size,
            binary_close=binary_close,
            val_for_max_rad=val_for_max_rad,
            max_rad_km=max_rad_km,
            core_val=core_val,
            nosfcecho=nosfcecho,
            weakecho=weakecho,
            bkgd_val=bkgd_val,
            feat_val=feat_val,
        )

        # get overestimate field
        if overest_field is None:
            over_field = ze + estimate_offset
        elif overest_field is not None:
            over_field = np.ma.copy(grid.fields[overest_field]["data"][0, :, :])

        # run algorithm to get feature detection underestimate
        _, _, feature_over = _feature_detection(
            over_field,
            dx,
            dy,
            always_core_thres=always_core_thres,
            bkg_rad_km=bkg_rad_km,
            use_cosine=use_cosine,
            max_diff=max_diff,
            zero_diff_cos_val=zero_diff_cos_val,
            scalar_diff=scalar_diff,
            use_addition=use_addition,
            calc_thres=calc_thres,
            weak_echo_thres=weak_echo_thres,
            min_val_used=min_val_used,
            dB_averaging=dB_averaging,
            remove_small_objects=remove_small_objects,
            min_km2_size=min_km2_size,
            binary_close=binary_close,
            val_for_max_rad=val_for_max_rad,
            max_rad_km=max_rad_km,
            core_val=core_val,
            nosfcecho=nosfcecho,
            weakecho=weakecho,
            bkgd_val=bkgd_val,
            feat_val=feat_val,
        )

        # save into dictionaries
        feature_dict["feature_under"] = {
            "data": feature_under[None, :, :],
            "standard_name": "feature_detection_under",
            "long_name": "Feature Detection Underestimate",
            "valid_min": 0,
            "valid_max": 3,
            "comment_1": (
                f"{nosfcecho} = No surface echo/Undefined, {bkgd_val} = Background echo, {feat_val} = Features, {weakecho} = weak echo"
            ),
        }

        feature_dict["feature_over"] = {
            "data": feature_over[None, :, :],
            "standard_name": "feature_detection_over",
            "long_name": "Feature Detection Overestimate",
            "valid_min": 0,
            "valid_max": 3,
            "comment_1": (
                f"{nosfcecho} = No surface echo/Undefined, {bkgd_val} = Background echo, {feat_val} = Features, {weakecho} = weak echo"
            ),
        }

    return feature_dict


def hydroclass_semisupervised(
    radar,
    hydro_names=("AG", "CR", "LR", "RP", "RN", "VI", "WS", "MH", "IH/HDG"),
    var_names=("Zh", "ZDR", "KDP", "RhoHV", "relH"),
    mass_centers=None,
    weights=np.array([1.0, 1.0, 1.0, 0.75, 0.5]),
    value=50.0,
    lapse_rate=-6.5,
    refl_field=None,
    zdr_field=None,
    rhv_field=None,
    kdp_field=None,
    temp_field=None,
    iso0_field=None,
    hydro_field=None,
    entropy_field=None,
    radar_freq=None,
    temp_ref="temperature",
    compute_entropy=False,
    output_distances=False,
    vectorize=False,
):
    """
    Classifies precipitation echoes into hydrometeor types.

    The `hydroclass_semisupervised` function classifies precipitation echoes in the polarimetric radar data
    into 9 hydrometeor types using a semi-supervised approach (Besic et al., 2016).

    Parameters
    ----------
    radar : radar
        Radar object.
    hydro_names : array of str
        name of the types of hydrometeors
    var_names : array of str
        name of the variables
    mass_centers : ndarray 2D, optional
        The centroids for each variable and hydrometeor class in (nclasses,
        nvariables).
    weights : ndarray 1D, optional
        The weight given to each variable. Ordered by [dBZ, ZDR, KDP, RhoHV,
        H_ISO0]
    value : float
        The value controlling the rate of decay in the distance transformation
    lapse_rate : float
        The decrease in temperature for each vertical km [deg/km]
    refl_field, zdr_field, rhv_field, kdp_field, temp_field, iso0_field : str
        Inputs. Field names within the radar object which represent the
        horizonal reflectivity, the differential reflectivity, the copolar
        correlation coefficient, the specific differential phase, the
        temperature (in deg celsius) and the height respect to the iso0 fields.
        A value of None for any of these parameters will use the default field
        name as defined in the Py-ART configuration file.
    hydro_field : str
        Output. Field name which represents the hydrometeor class field.
        A value of None will use the default field name as defined in the
        Py-ART configuration file.
    entropy_field : str
        Output. Field name which represents the entropy class field.
        A value of None will use the default field name as defined in the
        Py-ART configuration file.
    radar_freq : str, optional
        Radar frequency in Hertz (Hz) used for classification.
        This parameter will be ignored, if the radar object has frequency information.
    temp_ref : str
        the field use as reference for temperature. Can be either temperature
        or height_over_iso0
    compute_entropy : bool
        If true, the entropy is computed
    output_distances : bool
        If true, the normalized distances to the centroids for each
        hydrometeor are provided as output
    vectorize : bool
        If true, a vectorized version of the class assignation is going to be
        used

    Returns
    -------
    fields_dict : dict
        Dictionary containing the retrieved fields

    The output directionary field_dict has the following keys:

    hydro : dict
        Hydrometeor classification.
            - 0: Not classified
            - 1: Aggregates
            - 2: Ice crystals
            - 3: Light rain
            - 4: Rimed particles
            - 5: Rain
            - 6: Vertically oriented ice
            - 7: Wet snow
            - 8: Melting hail
            - 9: Dry hail or high-density graupel

    if compute_entropy is True:
    entropy : dict
        Shannon entropy of the hydrometeor demixing

    if output_distances is True:
    propX : dict
        Proportion of a given hydrometeor class in the polarimetric
        decomposition of a radar volume

    References
    ----------
    Besic, N., Figueras i Ventura, J., Grazioli, J., Gabella, M., Germann, U.,
    and Berne, A.: Hydrometeor classification through statistical clustering
    of polarimetric radar measurements: a semi-supervised approach,
    Atmos. Meas. Tech., 9, 4425-4445, doi:10.5194/amt-9-4425-2016, 2016

    Besic, N., Gehring, J., Praz, C., Figueras i Ventura, J., Grazioli, J., Gabella,
    M., Germann, U., and Berne, A.: Unraveling hydrometeor mixtures in polarimetric
    radar measurements, Atmos. Meas. Tech., 11, 4847–4866,
    https://doi.org/10.5194/amt-11-4847-2018, 2018.

    Usage
    -----
    .. code-block:: python
        hydro_class = pyart.retrieve.hydroclass_semisupervised(
            radar,
            refl_field="corrected_reflectivity",
            zdr_field="corrected_differential_reflectivity",
            kdp_field="specific_differential_phase",
            rhv_field="uncorrected_cross_correlation_ratio",
            temp_field="temperature",
        )

    Notes
    -----
    The default hydrometeor classification is valid for C-band radars. For X-band radars,
    if frequency information is not present in the `radar.instrument_parameters`, the user-supplied
    `radar_freq` will be used with a warning. If both `radar.instrument_parameters` and
    `radar_freq` parameter are missing, the algorithm defaults to the C band.

    If the radar frequency information is missing from the radar object, you can add it in
    `radar.instrument_parameters`, as follows:

    .. code-block:: python
        radar.instrument_parameters["frequency"] = {
            "long_name": "Radar frequency",
            "units": "Hz",
            "data": [9.2e9]
        }
    """

    # select the centroids as a function of frequency band
    if mass_centers is None:
        # assign coefficients according to radar frequency
        if radar.instrument_parameters and "frequency" in radar.instrument_parameters:
            frequency = radar.instrument_parameters["frequency"]["data"][0]
            mass_centers = _get_mass_centers(frequency)
            warn(f"Using radar frequency from instrument parameters: {frequency}")
        elif radar_freq is not None:
            mass_centers = _get_mass_centers(radar_freq)
            warn(
                f"Radar instrument parameters are empty. Using user-supplied radar frequency: {radar_freq}"
            )
        else:
            mass_centers = _mass_centers_table()["C"]
            warn(
                "Radar instrument parameters and radar_freq param are empty."
                "So frequency is unknown. Default coefficients for C band will be applied."
            )

    if hydro_field is None:
        hydro_field = get_field_name("radar_echo_classification")
    if compute_entropy:
        if entropy_field is None:
            entropy_field = get_field_name("hydroclass_entropy")

    # Get the data fields
    fields_dict = {}
    for var_name in var_names:
        if var_name == "Zh":
            if refl_field is None:
                refl_field = get_field_name("reflectivity")
            radar.check_field_exists(refl_field)
            fields_dict.update({var_name: deepcopy(radar.fields[refl_field]["data"])})
        elif var_name == "ZDR":
            if zdr_field is None:
                zdr_field = get_field_name("differential_reflectivity")
            radar.check_field_exists(zdr_field)
            fields_dict.update({var_name: deepcopy(radar.fields[zdr_field]["data"])})
        elif var_name == "KDP":
            if kdp_field is None:
                kdp_field = get_field_name("specific_differential_phase")
            radar.check_field_exists(kdp_field)
            fields_dict.update({var_name: deepcopy(radar.fields[kdp_field]["data"])})
        elif var_name == "RhoHV":
            if rhv_field is None:
                rhv_field = get_field_name("cross_correlation_ratio")
            radar.check_field_exists(rhv_field)
            fields_dict.update({var_name: deepcopy(radar.fields[rhv_field]["data"])})
        elif var_name == "relH":
            if temp_ref == "temperature":
                if temp_field is None:
                    temp_field = get_field_name("temperature")
                radar.check_field_exists(temp_field)
                # convert temp in relative height respect to iso0
                temp = deepcopy(radar.fields[temp_field]["data"])
                fields_dict.update({var_name: temp * (1000.0 / lapse_rate)})
            else:
                if iso0_field is None:
                    iso0_field = get_field_name("height_over_iso0")
                radar.check_field_exists(iso0_field)
                fields_dict.update(
                    {var_name: deepcopy(radar.fields[iso0_field]["data"])}
                )
        else:
            raise ValueError(
                "Variable "
                + var_name
                + " unknown. "
                + "Valid variable names for hydrometeor classification are: "
                + "relH, Zh, ZDR, KDP and RhoHV"
            )

    # standardize data and centroids
    mc_std = np.empty(np.shape(mass_centers), dtype=fields_dict[var_names[0]].dtype)
    for i, var_name in enumerate(var_names):
        mc_std[:, i] = _standardize(mass_centers[:, i], var_name)
        fields_dict[var_name] = _standardize(fields_dict[var_name], var_name)

    # if entropy has to be computed get transformation parameters
    t_vals = None
    if compute_entropy:
        t_vals = _compute_coeff_transform(mc_std, weights=weights, value=value)

    # assign to class
    if vectorize:
        hydroclass_data, entropy_data, prop_data = _assign_to_class_scan(
            fields_dict, mc_std, var_names=var_names, weights=weights, t_vals=t_vals
        )
    else:
        hydroclass_data, entropy_data, prop_data = _assign_to_class(
            fields_dict, mc_std, weights=weights, t_vals=t_vals
        )

    # prepare output fields
    fields_dict = dict()
    hydro = get_metadata(hydro_field)
    hydro["data"] = hydroclass_data
    hydro.update({"_FillValue": 0})
    labels = ["NC"]
    ticks = [1]
    boundaries = [-0.5, 1.5]
    for i, hydro_name in enumerate(hydro_names):
        labels.append(hydro_name)
        ticks.append(i + 1)
        boundaries.append(i + 1.5)
    hydro.update({"labels": labels, "ticks": ticks, "boundaries": boundaries})
    fields_dict.update({"hydro": hydro})

    if compute_entropy:
        entropy = get_metadata(entropy_field)
        entropy["data"] = entropy_data
        fields_dict.update({"entropy": entropy})

        if output_distances:
            for field_name in hydro_names:
                field_name = "proportion_" + field_name
                prop = get_metadata(field_name)
                prop["data"] = prop_data[:, :, i]
                fields_dict.update({field_name: prop})

    return fields_dict


def _standardize(data, field_name, mx=None, mn=None):
    """
    Streches the radar data to -1 to 1 interval.

    Parameters
    ----------
    data : array
        Radar field.
    field_name : str
        Type of field (relH, Zh, ZDR, KDP or RhoHV).
    mx, mn : floats or None, optional
        Data limits for array values.

    Returns
    -------
    field_std : dict
        Standardized radar data.

    """
    if field_name == "relH":
        field_std = 2.0 / (1.0 + np.ma.exp(-0.005 * data)) - 1.0
        return field_std

    if (mx is None) or (mn is None):
        dlimits_dict = _data_limits_table()
        if field_name not in dlimits_dict:
            raise ValueError(
                "Field "
                + field_name
                + " unknown. "
                + "Valid field names for standardizing are: "
                + "relH, Zh, ZDR, KDP and RhoHV"
            )
        mx, mn = dlimits_dict[field_name]

    if field_name == "KDP":
        data[data < -0.5] = -0.5
        data = 10.0 * np.ma.log10(data + 0.6)
    elif field_name == "RhoHV":
        # avoid infinite result
        data[data > 1.0] = 1.0
        data = 10.0 * np.ma.log10(1.0000000000001 - data)

    mask = np.ma.getmaskarray(data)
    field_std = 2.0 * (data - mn) / (mx - mn) - 1.0
    field_std[data < mn] = -1.0
    field_std[data > mx] = 1.0
    field_std[mask] = np.ma.masked

    return field_std


def _assign_to_class(
    fields_dict,
    mass_centers,
    var_names=("Zh", "ZDR", "KDP", "RhoHV", "relH"),
    weights=np.array([1.0, 1.0, 1.0, 0.75, 0.5]),
    t_vals=None,
):
    """
    Assigns an hydrometeor class to a radar range bin computing
    the distance between the radar variables an a centroid.

    Parameters
    ----------
    fields_dict : dict
        Dictionary containg the variables used for assigment normalized to
        [-1, 1] values
    mass_centers : matrix
        centroids normalized to [-1, 1] values (nclasses, nvariables)
    var_names : array of str
        Name of the variables
    weights : array
        optional. The weight given to each variable (nvariables)
    t_vals : array
        transformation values for the distance to centroids (nclasses)

    Returns
    -------
    hydroclass : int array
        the index corresponding to the assigned class
    entropy : float array
        the entropy
    t_dist : float matrix
        if entropy is computed, the transformed distances of each class
        (proxy for proportions of each hydrometeor) (nrays, nbins, nclasses)

    """
    # prepare data
    nrays = fields_dict[var_names[0]].shape[0]
    nbins = fields_dict[var_names[0]].shape[1]
    nclasses = mass_centers.shape[0]
    nvariables = mass_centers.shape[1]
    dtype = fields_dict[var_names[0]].dtype

    hydroclass = np.ma.empty((nrays, nbins), dtype=np.uint8)
    entropy = None
    t_dist = None
    if t_vals is not None:
        entropy = np.ma.empty((nrays, nbins), dtype=dtype)
        t_dist = np.ma.masked_all((nrays, nbins, nclasses), dtype=dtype)

    for ray in range(nrays):
        data = []
        for var_name in var_names:
            data.append(fields_dict[var_name][ray, :])
        data = np.ma.array(data, dtype=dtype)
        weights_mat = np.broadcast_to(
            weights.reshape(nvariables, 1), (nvariables, nbins)
        )
        dist = np.ma.zeros((nclasses, nbins), dtype=dtype)

        # compute distance: masked entries will not contribute to the distance
        mask = np.ma.getmaskarray(fields_dict[var_names[0]][ray, :])
        for i in range(nclasses):
            centroids_class = mass_centers[i, :]
            centroids_class = np.broadcast_to(
                centroids_class.reshape(nvariables, 1), (nvariables, nbins)
            )
            dist_ray = np.ma.sqrt(
                np.ma.sum(((centroids_class - data) ** 2.0) * weights_mat, axis=0)
            )
            dist_ray[mask] = np.ma.masked
            dist[i, :] = dist_ray

        # Get hydrometeor class
        class_vec = dist.argsort(axis=0, fill_value=10e40)
        hydroclass_ray = (class_vec[0, :] + 1).astype(np.uint8)
        hydroclass_ray[mask] = 0
        hydroclass[ray, :] = hydroclass_ray

        if t_vals is None:
            continue

        # Transform the distance using the coefficient of the dominant class
        t_vals_ray = np.ma.masked_where(mask, t_vals[class_vec[0, :]])
        t_vals_ray = ma_broadcast_to(t_vals_ray.reshape(1, nbins), (nclasses, nbins))
        t_dist_ray = np.ma.exp(-t_vals_ray * dist)

        # set transformed distances to a value between 0 and 1
        dist_total = np.ma.sum(t_dist_ray, axis=0)
        dist_total = ma_broadcast_to(dist_total.reshape(1, nbins), (nclasses, nbins))
        t_dist_ray /= dist_total

        # Compute entropy
        entropy_ray = -np.ma.sum(
            t_dist_ray * np.ma.log(t_dist_ray) / np.ma.log(nclasses), axis=0
        )
        entropy_ray[mask] = np.ma.masked
        entropy[ray, :] = entropy_ray

        t_dist[ray, :, :] = np.ma.transpose(t_dist_ray)

    if t_vals is not None:
        t_dist *= 100.0

    return hydroclass, entropy, t_dist


def _assign_to_class_scan(
    fields_dict,
    mass_centers,
    var_names=("Zh", "ZDR", "KDP", "RhoHV", "relH"),
    weights=np.array([1.0, 1.0, 1.0, 0.75, 0.5]),
    t_vals=None,
):
    """
    assigns an hydrometeor class to a radar range bin computing
    the distance between the radar variables an a centroid.
    Computes the entire radar volume at once

    Parameters
    ----------
    fields_dict : dict
        Dictionary containg the variables used for assigment normalized to
        [-1, 1] values
    mass_centers : matrix
        centroids normalized to [-1, 1] values
    var_names : array of str
        Name of the variables
    weights : array
        optional. The weight given to each variable
    t_vals : matrix
        transformation values for the distance to centroids
        (nclasses, nvariables)

    Returns
    -------
    hydroclass : int array
        the index corresponding to the assigned class
    entropy : float array
        the entropy
    t_dist : float matrix
        if entropy is computed, the transformed distances of each class
        (proxy for proportions of each hydrometeor) (nrays, nbins, nclasses)

    """
    # prepare data
    nrays = fields_dict[var_names[0]].shape[0]
    nbins = fields_dict[var_names[0]].shape[1]
    nclasses = mass_centers.shape[0]
    nvariables = mass_centers.shape[1]
    dtype = fields_dict[var_names[0]].dtype

    data = []
    for var_name in var_names:
        data.append(fields_dict[var_name])
    data = np.ma.array(data, dtype=dtype)
    weights_mat = np.broadcast_to(
        weights.reshape(nvariables, 1, 1), (nvariables, nrays, nbins)
    )

    # compute distance: masked entries will not contribute to the distance
    mask = np.ma.getmaskarray(fields_dict[var_names[0]])
    dist = np.ma.zeros((nrays, nbins, nclasses), dtype=dtype)
    t_dist = None
    entropy = None
    for i in range(nclasses):
        centroids_class = mass_centers[i, :]
        centroids_class = np.broadcast_to(
            centroids_class.reshape(nvariables, 1, 1), (nvariables, nrays, nbins)
        )
        dist_aux = np.ma.sqrt(
            np.ma.sum(((centroids_class - data) ** 2.0) * weights_mat, axis=0)
        )
        dist_aux[mask] = np.ma.masked
        dist[:, :, i] = dist_aux

    del data
    del weights_mat

    # Get hydrometeor class
    class_vec = dist.argsort(axis=-1, fill_value=10e40)
    hydroclass = np.ma.asarray(class_vec[:, :, 0] + 1, dtype=np.uint8)
    hydroclass[mask] = 0

    if t_vals is not None:
        # Transform the distance using the coefficient of the dominant class
        t_vals_aux = np.ma.masked_where(mask, t_vals[class_vec[:, :, 0]])
        t_vals_aux = ma_broadcast_to(
            t_vals_aux.reshape(nrays, nbins, 1), (nrays, nbins, nclasses)
        )
        t_dist = np.ma.exp(-t_vals_aux * dist)
        del t_vals_aux

        # set distance to a value between 0 and 1
        dist_total = np.ma.sum(t_dist, axis=-1)
        dist_total = ma_broadcast_to(
            dist_total.reshape(nrays, nbins, 1), (nrays, nbins, nclasses)
        )
        t_dist /= dist_total
        del dist_total

        # compute entroy
        entropy = -np.ma.sum(t_dist * np.ma.log(t_dist) / np.ma.log(nclasses), axis=-1)
        entropy[mask] = np.ma.masked

        t_dist *= 100.0

    return hydroclass, entropy, t_dist


def _get_mass_centers(freq):
    """
    Get mass centers for a particular frequency.

    Parameters
    ----------
    freq : float
        Radar frequency [Hz].

    Returns
    -------
    mass_centers : ndarray 2D
        The centroids for each variable and hydrometeor class in (nclasses,
        nvariables).

    """
    mass_centers_dict = _mass_centers_table()

    freq_band = get_freq_band(freq)
    if (freq_band is not None) and (freq_band in mass_centers_dict):
        return mass_centers_dict[freq_band]

    if freq < 4e9:
        freq_band_aux = "C"
    elif freq > 12e9:
        freq_band_aux = "X"

    mass_centers = mass_centers_dict[freq_band_aux]
    warn(
        "Radar frequency out of range. "
        + "Centroids only valid for C or X band. "
        + freq_band_aux
        + " band centroids will be applied"
    )

    return mass_centers


def _mass_centers_table():
    """
    Defines the mass centers look up table for each frequency band.

    Returns
    -------
    mass_centers_dict : dict
        A dictionary with the mass centers for each frequency band.

    """
    nclasses = 9
    nvariables = 5
    mass_centers_c = np.zeros((nclasses, nvariables))
    mass_centers_x = np.zeros((nclasses, nvariables))

    mass_centers_dict = dict()
    # C-band centroids derived for MeteoSwiss Albis radar
    #                       Zh        ZDR     kdp   RhoHV    delta_Z
    mass_centers_c[0, :] = [13.5829, 0.4063, 0.0497, 0.9868, 1330.3]  # DS
    mass_centers_c[1, :] = [02.8453, 0.2457, 0.0000, 0.9798, 0653.8]  # CR
    mass_centers_c[2, :] = [07.6597, 0.2180, 0.0019, 0.9799, -1426.5]  # LR
    mass_centers_c[3, :] = [31.6815, 0.3926, 0.0828, 0.9978, 0535.3]  # GR
    mass_centers_c[4, :] = [39.4703, 1.0734, 0.4919, 0.9876, -1036.3]  # RN
    mass_centers_c[5, :] = [04.8267, -0.5690, 0.0000, 0.9691, 0869.8]  # VI
    mass_centers_c[6, :] = [30.8613, 0.9819, 0.1998, 0.9845, -0066.1]  # WS
    mass_centers_c[7, :] = [52.3969, 2.1094, 2.4675, 0.9730, -1550.2]  # MH
    mass_centers_c[8, :] = [50.6186, -0.0649, 0.0946, 0.9904, 1179.9]  # IH/HDG

    mass_centers_dict.update({"C": mass_centers_c})

    # X-band centroids derived for MeteoSwiss DX50 radar
    #                       Zh        ZDR     kdp    RhoHV   delta_Z
    mass_centers_x[0, :] = [19.0770, 0.4139, 0.0099, 0.9841, 1061.7]  # DS
    mass_centers_x[1, :] = [03.9877, 0.5040, 0.0000, 0.9642, 0856.6]  # CR
    mass_centers_x[2, :] = [20.7982, 0.3177, 0.0004, 0.9858, -1375.1]  # LR
    mass_centers_x[3, :] = [34.7124, -0.3748, 0.0988, 0.9828, 1224.2]  # GR
    mass_centers_x[4, :] = [33.0134, 0.6614, 0.0819, 0.9802, -1169.8]  # RN
    mass_centers_x[5, :] = [08.2610, -0.4681, 0.0000, 0.9722, 1100.7]  # VI
    mass_centers_x[6, :] = [35.1801, 1.2830, 0.1322, 0.9162, -0159.8]  # WS
    mass_centers_x[7, :] = [52.4539, 2.3714, 1.1120, 0.9382, -1618.5]  # MH
    mass_centers_x[8, :] = [44.2216, -0.3419, 0.0687, 0.9683, 1272.7]  # IH/HDG

    mass_centers_dict.update({"X": mass_centers_x})

    return mass_centers_dict


def _data_limits_table():
    """
    Defines the data limits used in the standardization.

    Returns
    -------
    dlimits_dict : dict
        A dictionary with the limits for each variable.

    """
    dlimits_dict = dict()
    dlimits_dict.update({"Zh": (60.0, -10.0)})
    dlimits_dict.update({"ZDR": (5.0, -1.5)})
    dlimits_dict.update({"KDP": (7.0, -10.0)})
    dlimits_dict.update({"RhoHV": (-5.23, -50.0)})
    dlimits_dict.update({"RelH": (5000.0, -5000.0)})

    return dlimits_dict


def get_freq_band(freq):
    """
    Returns the frequency band name (S, C, X, ...).

    Parameters
    ----------
    freq : float
        Radar frequency [Hz].

    Returns
    -------
    freq_band : str
        Frequency band name.

    """
    if freq >= 2e9 and freq < 4e9:
        return "S"
    if freq >= 4e9 and freq < 8e9:
        return "C"
    if freq >= 8e9 and freq <= 12e9:
        return "X"

    warn("Unknown frequency band")

    return None


def conv_strat_raut(
    grid,
    refl_field,
    cappi_level=0,
    zr_a=200,
    zr_b=1.6,
    core_wt_threshold=5,
    conv_wt_threshold=1.5,
    conv_scale_km=25,
    min_reflectivity=5,
    conv_min_refl=25,
    conv_core_threshold=42,
    override_checks=False,
):
    """
    A computationally efficient method to classify radar echoes into convective cores, mixed convection,
    and stratiform regions for gridded radar reflectivity field.

    This function uses à trous wavelet transform (ATWT) for multiresolution (i.e. scale) analysis of radar field,
    focusing on precipitation structure over reflectivity thresholds for robust echo classification (Raut et al 2008, 2020).

    Parameters
    ----------
    grid : PyART Grid
        Grid object containing radar data.
    refl_field : str
        Field name for reflectivity data in the Py-ART grid object.
    zr_a : float, optional
        Coefficient 'a' in the Z-R relationship Z = a*R^b for reflectivity to rain rate conversion.
        The algorithm is not sensitive to precise values of 'zr_a' and 'zr_b'; however,
        they must be adjusted based on the type of radar used.
        Default is 200.
    zr_b : float, optional
        Coefficient 'b' in the Z-R relationship Z = a*R^b. Default is 1.6.
    core_wt_threshold : float, optional
        Threshold for wavelet components to separate convective cores from mix-intermediate type.
        Default is 5. Recommended values are between 4 and 6.
    conv_wt_threshold : float, optional
        Threshold for significant wavelet components to separate all convection from stratiform.
        Default is 1.5. Recommended values are between 1 and 2.
    conv_scale_km : float, optional
        Approximate scale break (in km) between convective and stratiform scales.
        Scale break may vary over different regions and seasons
        (Refere to Raut et al 2018 for more discussion on scale-breaks). Note that the
        algorithm is insensitive to small variations in the scale break due to the
        dyadic nature of the scaling. The actual scale break used in the calculation of wavelets
        is returned in the output dictionary by parameter `scale_break_used`.
        Default is 25 km. Recommended values are between 16 and 32 km.
    min_reflectivity : float, optional
        Minimum reflectivity threshold. Reflectivities below this value are not classified.
        Default is 5 dBZ. This value must be greater than or equal to '0'.
    conv_min_refl : float, optional
        Reflectivity values lower than this threshold will be always considered as non-convective.
        Default is 25 dBZ. Recommended values are between 25 and 30 dBZ.
    conv_core_threshold : float, optional
        Reflectivities above this threshold are classified as convective cores if wavelet components are significant (See: conv_wt_threshold).
        Default is 42 dBZ.
        Recommended value must be is greater than or equal to 40 dBZ. The algorithm is not sensitive to this value.
    override_checks : bool, optional
        If set to True, the function will bypass the sanity checks for above parameter values.
        This allows the user to use custom values for parameters, even if they fall outside
        the recommended ranges. The default is False.

    Returns
    -------

    dict :
    A dictionary structured as a Py-ART grid field, suitable for adding to a Py-ART Grid object. The dictionary
    contains the classification data and associated metadata. The classification categories are as follows:
        - 3: Convective Cores: associated with strong updrafts and active collision-coalescence.
        - 2: Mixed-Intermediate: capturing a wide range of convective activities, excluding the convective cores.
        - 1: Stratiform: remaining areas with more uniform and less intense precipitation.
        - 0: Unclassified: for reflectivity below the minimum threshold.


    References
    ----------
    Raut, B. A., Karekar, R. N., & Puranik, D. M. (2008). Wavelet-based technique to extract convective clouds from
    infrared satellite images. IEEE Geosci. Remote Sens. Lett., 5(3), 328-330.

    Raut, B. A., Seed, A. W., Reeder, M. J., & Jakob, C. (2018). A multiplicative cascade model for high‐resolution
    space‐time downscaling of rainfall. J. Geophys. Res. Atmos., 123(4), 2050-2067.

    Raut, B. A., Louf, V., Gayatri, K., Murugavel, P., Konwar, M., & Prabhakaran, T. (2020). A multiresolution technique
    for the classification of precipitation echoes in radar data. IEEE Trans. Geosci. Remote Sens., 58(8), 5409-5415.
    """

    # Check if the grid is a Py-ART Grid object
    if not isinstance(grid, Grid):
        raise TypeError("The 'grid' is not a Py-ART Grid object.")

    # Check if dx and dy are the same, and warn if not
    dx = grid.x["data"][1] - grid.x["data"][0]
    dy = grid.y["data"][1] - grid.y["data"][0]
    if dx != dy:
        warn(
            "Warning: Grid resolution `dx` and `dy` should be comparable for correct results.",
            UserWarning,
        )

    # Compute scale break (dyadic) here to paas it on as parameter to user dictionary
    scale_break = calc_scale_break(res_meters=dx, conv_scale_km=conv_scale_km)

    # From dyadic scale to km
    scale_break_km = (2 ** (scale_break - 1)) * dx / 1000

    # Sanity checks for parameters if override_checks is False
    if not override_checks:
        conv_core_threshold = max(
            40, conv_core_threshold
        )  # Ensure conv_core_threshold is at least 40 dBZ
        core_wt_threshold = max(
            4, min(core_wt_threshold, 6)
        )  # core_wt_threshold should be between 4 and 6
        conv_wt_threshold = max(
            1, min(conv_wt_threshold, 2)
        )  # conv_wt_threshold should be between 1 and 2
        conv_scale_km = max(
            16, min(conv_scale_km, 32)
        )  # conv_scale_km should be between 15 and 30 km
        min_reflectivity = max(
            0, min_reflectivity
        )  # min_reflectivity should be non-negative
        conv_min_refl = max(
            25, min(conv_min_refl, 30)
        )  # conv_min_refl should be between 25 and 30 dBZ

    # Call the actual wavelet_relass function to obtain radar echo classificatino
    reclass = wavelet_reclass(
        grid,
        refl_field,
        cappi_level,
        zr_a,
        zr_b,
        core_wt_threshold=core_wt_threshold,
        conv_wt_threshold=conv_wt_threshold,
        scale_break=scale_break,
        min_reflectivity=min_reflectivity,
        conv_min_refl=conv_min_refl,
        conv_core_threshold=conv_core_threshold,
    )

    reclass = np.expand_dims(reclass, axis=0)

    # put data into a dictionary to be added as a field
    reclass_dict = {
        "wt_reclass": {
            "data": reclass,
            "standard_name": "wavelet_echo_class",
            "long_name": "Wavelet-based multiresolution radar echo classification",
            "valid_min": 0,
            "valid_max": 3,
            "classification_description": "0: Unclassified, 1: Stratiform, 2: Mixed-Intermediate, 3: Convective Cores",
            "parameters": {
                "refl_field": refl_field,
                "cappi_level": cappi_level,
                "zr_a": zr_a,
                "zr_b": zr_b,
                "core_wt_threshold": core_wt_threshold,
                "conv_wt_threshold": conv_wt_threshold,
                "conv_scale_km": conv_scale_km,
                "scale_break_used": int(scale_break_km),
                "min_reflectivity": min_reflectivity,
                "conv_min_refl": conv_min_refl,
                "conv_core_threshold": conv_core_threshold,
            },
        }
    }

    return reclass_dict


def _compute_coeff_transform(
    mass_centers, weights=np.array([1.0, 1.0, 1.0, 0.75, 0.5]), value=50.0
):
    """
    get the transformation coefficients

    Parameters
    ----------
    mass_centers : ndarray 2D
        The centroids for each class and variable (nclasses, nvariables)
    weights : array
        optional. The weight given to each variable (nvariables)
    value : float
        parameter controlling the rate of decay of the distance transformation

    Returns
    -------
    t_vals : ndarray 1D
        The coefficients used to transform the distances to each centroid for
        each class (nclasses)

    """
    nclasses, nvariables = np.shape(mass_centers)
    t_vals = np.empty((nclasses, nclasses), dtype=mass_centers.dtype)
    for i in range(nclasses):
        weights_mat = np.broadcast_to(
            weights.reshape(1, nvariables), (nclasses, nvariables)
        )
        centroids_class = mass_centers[i, :]
        centroids_class = np.broadcast_to(
            centroids_class.reshape(1, nvariables), (nclasses, nvariables)
        )
        t_vals[i, :] = np.sqrt(
            np.sum(
                weights_mat * np.power(np.abs(centroids_class - mass_centers), 2.0),
                axis=1,
            )
        )

    # pick the second lowest value (the first is 0)
    t_vals = np.sort(t_vals, axis=-1)[:, 1]
    t_vals = np.log(value) / t_vals

    return t_vals



"""
======================================
Map a single PPI sweep to a Cartesian grid using 2D weighting
======================================

Map the reflectivity field of a single PPI sweep from Antenna (polar) coordinates
to a Cartesian grid, while using a 2D weighting.
This solution is valid only for this case (single PPI sweep), yet it is a useful one
(an arguably global solution since it overlooks the z-dimension grid limits).
The exclusion of the z-dimension from the RoI and weighting calculations results in
minor errors, especially considering the high co-variance of neighboring radar
volumes and the typically small Cartesian grid separation. Errors are effectively
negligible at low elevation angles common to PPI sweeps.

"""

print(__doc__)

# =====================
# Author: Israel Silber
# =====================

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
from open_radar_data import DATASETS

import pyart

# file, and fields
# =======================
file_name = DATASETS.fetch("example_plot_ppi_single_sweep.nc")
processed_field = "reflectivity_at_cor"

# load file
# =======================
radar_obj = pyart.io.read(file_name)
print(f"Total sweeps = {radar_obj.nsweeps}")

# Plot polar coordinate (raw) data
# =======================
print(
    "===\n\nnow displaying Ze data for 2nd and 4th sweeps in polar coordinates\n"
    "(both approaching the 40 km range denoted by the purple ring)\n\n==="
)
fig = plt.figure(figsize=(12, 6), tight_layout=True)
fig.suptitle("polar")
for sweep, ax_ind in zip([1, 3], range(2)):
    ax = fig.add_subplot(1, 2, ax_ind + 1, projection=ccrs.Mercator())
    sweep_obj = radar_obj.extract_sweeps((sweep,))
    display = pyart.graph.RadarDisplay(sweep_obj)
    display.plot(
        processed_field,
        sweep=0,
        ax=ax,
    )
    display.plot_range_rings([10, 20, 30])
    display.plot_range_rings([40], col="purple")
plt.show()

print(
    "===\n\nnow displaying gridded Ze data for 2nd and 4th (final) sweeps; note the "
    "truncated max range in the case of the 4th sweep\n\n==="
)
fig2 = plt.figure(figsize=(12, 6), tight_layout=True)
fig2.suptitle("Cartesian gridded")
for sweep, ax_ind in zip([1, 3], range(2)):
    sweep_obj = radar_obj.extract_sweeps((sweep,))
    grid = pyart.map.grid_from_radars(
        (sweep_obj,),
        grid_shape=(1, 1601, 1601),
        grid_limits=((0, 10000.0), [-40000, 40000], [-40000, 40000]),
        fields=[processed_field],
    )
    ax = fig2.add_subplot(1, 2, ax_ind + 1)
    ax.imshow(
        grid.fields[processed_field]["data"][0],
        origin="lower",
        extent=(-40, 40, -40, 40),
    )
    ax.set_title(f"sweep #{sweep + 1}")
plt.show()

print(
    "===\n\nUsing 2D weighting by "
    "setting h_factor and dist_factor z component to 0.0, the max range looks OK now\n\n==="
)
fig2 = plt.figure(figsize=(12, 6), tight_layout=True)
fig2.suptitle("Cartesian gridded")
for sweep, ax_ind in zip([1, 3], range(2)):
    sweep_obj = radar_obj.extract_sweeps((sweep,))
    grid = pyart.map.grid_from_radars(
        (sweep_obj,),
        grid_shape=(1, 1601, 1601),
        grid_limits=((0, 10000.0), [-40000, 40000], [-40000, 40000]),
        fields=[processed_field],
        h_factor=(0.0, 1.0, 1.0),
        dist_factor=(0.0, 1.0, 1.0),
    )
    ax = fig2.add_subplot(1, 2, ax_ind + 1)
    ax.imshow(
        grid.fields[processed_field]["data"][0],
        origin="lower",
        extent=(-40, 40, -40, 40),
    )
    ax.set_title(f"sweep #{sweep + 1}")
plt.show()



"""
====================
Compare PPI vs CAPPI
====================

This example demonstrates how to create and compare PPI (Plan Position Indicator)
and CAPPI (Constant Altitude Plan Position Indicator) plots using radar data.

In this example, we load sample radar data, create a CAPPI at 2,000 meters
for the 'reflectivity' field, and then plot both the PPI and CAPPI for comparison.

"""

print(__doc__)

# Author: Hamid Ali Syed (syed44@purdue.edu)
# License: BSD 3 clause

import matplotlib.pyplot as plt
from open_radar_data import DATASETS

import pyart

# Load the sample radar data
file = DATASETS.fetch("RAW_NA_000_125_20080411190016")
radar = pyart.io.read(file)

# Apply gate filtering to exclude unwanted data
gatefilter = pyart.filters.GateFilter(radar)
gatefilter.exclude_transition()

# Create CAPPI at 2,000 meters for the 'reflectivity' field
cappi = pyart.retrieve.create_cappi(
    radar, fields=["reflectivity"], height=2000, gatefilter=gatefilter
)

# Create RadarMapDisplay objects for both PPI and CAPPI
radar_display = pyart.graph.RadarMapDisplay(radar)
cappi_display = pyart.graph.RadarMapDisplay(cappi)

# Plotting the PPI and CAPPI for comparison
fig, ax = plt.subplots(1, 2, figsize=(13, 5))

# Plot PPI for 'reflectivity' field
radar_display.plot_ppi("reflectivity", vmin=-10, vmax=60, ax=ax[0])
ax[0].set_title("PPI Reflectivity")

# Plot CAPPI for 'reflectivity' field
cappi_display.plot_ppi("reflectivity", vmin=-10, vmax=60, ax=ax[1])
ax[1].set_title("CAPPI Reflectivity at 2000 meters")

# Show the plots
plt.show()



import numpy as np
from open_radar_data import DATASETS

import pyart
from pyart.retrieve import create_cappi


def test_create_cappi():
    # Load radar data
    file = DATASETS.fetch("RAW_NA_000_125_20080411190016")
    radar = pyart.io.read(file)

    # Create CAPPI at 10000 meters for the 'reflectivity' field
    cappi = create_cappi(radar, fields=["reflectivity"], height=10000)

    # Retrieve the 'reflectivity' field from the generated CAPPI
    reflectivity_cappi = cappi.fields["reflectivity"]

    # Test 1: Check the shape of the reflectivity CAPPI data
    expected_shape = (360, 992)  # As per the sample data provided
    assert (
        reflectivity_cappi["data"].shape == expected_shape
    ), "Shape mismatch in CAPPI data"

    # Test 2: Check the units of the reflectivity CAPPI
    assert (
        reflectivity_cappi["units"] == "dBZ"
    ), "Incorrect units for CAPPI reflectivity"

    # Test 3: Check that the elevation data is correctly set to zero
    assert np.all(
        cappi.elevation["data"] == 0
    ), "Elevation data should be all zeros in CAPPI"

    # Test 4: Verify the fill value
    assert (
        reflectivity_cappi["_FillValue"] == -9999.0
    ), "Incorrect fill value in CAPPI reflectivity"

    # Test 5: Check the long name and comment
    assert (
        reflectivity_cappi["long_name"] == "CAPPI reflectivity at 10000 meters"
    ), "Incorrect long name"
    assert (
        reflectivity_cappi["comment"]
        == "CAPPI reflectivity calculated at a height of 10000 meters"
    ), "Incorrect comment"



""" Unit Tests for Py-ART's retrieve/srv.py module. """

import numpy as np
import pytest

import pyart


def test_storm_relative_velocity():
    # Test all sweeps
    radar = pyart.io.read_nexrad_archive(
        "s3://noaa-nexrad-level2/2022/03/31/KDGX/KDGX20220331_012808_V06"
    )
    sr_data = pyart.retrieve.storm_relative_velocity(radar, direction="NE", speed=20.0)
    test_data = [2.276456117630005, 4.776455879211426, 3.276456117630005]
    np.testing.assert_almost_equal(sr_data[-1][0:3].tolist(), test_data)

    # Test one sweep
    radar_sweep = radar.extract_sweeps([21])
    sr_one_sweep = pyart.retrieve.storm_relative_velocity(
        radar_sweep, direction="NE", speed=20.0
    )
    one_sweep_data = [0.1278250813484192, 4.6278252601623535, 4.6278252601623535]
    np.testing.assert_almost_equal(sr_one_sweep[0][0:3].tolist(), one_sweep_data)

    # Test missing parameters
    pytest.raises(ValueError, pyart.retrieve.storm_relative_velocity, radar)
    pytest.raises(
        ValueError, pyart.retrieve.storm_relative_velocity, radar, u=14.14, v=None
    )



"""
===========================================
Create a 3 panel plot using GridMapDisplay
===========================================

An example that creates a 3 panel plot of a PPI, latitude slice,
and longitude slice using xarray and a cartopy background.

"""

print(__doc__)

# Author: Jason Hemedinger
# License: BSD 3 clause

import cartopy.crs as ccrs
import matplotlib.pyplot as plt

import pyart
from pyart.testing import get_test_data

# Read in the gridded file, create GridMapDisplay object
filename = get_test_data("20110520100000_nexrad_grid.nc")
radar = pyart.io.read_grid(filename)
display = pyart.graph.GridMapDisplay(radar)

# Setting projection, figure size, and panel sizes.
projection = ccrs.PlateCarree()

fig = plt.figure(figsize=[15, 7])

map_panel_axes = [0.05, 0.05, 0.4, 0.80]
x_cut_panel_axes = [0.55, 0.10, 0.4, 0.25]
y_cut_panel_axes = [0.55, 0.50, 0.4, 0.25]

# Set parameters.
level = 1
vmin = -8
vmax = 64
lat = 36.5
lon = -97.7

# Panel 1: PPI plot of the second tilt.
ax1 = fig.add_axes(map_panel_axes, projection=projection)
display.plot_grid(
    "REF",
    1,
    vmin=vmin,
    vmax=vmax,
    ax=ax1,
    projection=projection,
    cmap="HomeyerRainbow",
)
display.plot_crosshairs(lon=lon, lat=lat)

# Panel 2: longitude slice
ax2 = fig.add_axes(x_cut_panel_axes)
display.plot_longitude_slice(
    "REF", lon=lon, lat=lat, ax=ax2, vmin=vmin, vmax=vmax, cmap="HomeyerRainbow"
)

ax2.set_ylim([0, 15])
ax2.set_xlim([-50, 50])

# Panel 3: latitude slice
ax3 = fig.add_axes(y_cut_panel_axes)
display.plot_latitude_slice(
    "REF", lon=lon, lat=lat, ax=ax3, vmin=vmin, vmax=vmax, cmap="HomeyerRainbow"
)
ax3.set_ylim([0, 15])
ax3.set_xlim([-50, 50])

plt.show()
